/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.topology;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Logger;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration.ContainerState;
import org.ow2.petals.clientserverapi.jbi.messaging.registry.exception.RegistryException;
import org.ow2.petals.clientserverapi.server.exception.PetalsCompositeCtrlException;
import org.ow2.petals.clientserverapi.topology.exception.AddContainerException;
import org.ow2.petals.clientserverapi.topology.exception.AttachContainerException;
import org.ow2.petals.clientserverapi.topology.exception.DetachContainerException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistencyException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistencySubdomainException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistentContainerNameException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistentSubdomainNameException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistentSubdomainRegistryImplException;
import org.ow2.petals.clientserverapi.topology.exception.NoContainerConfigurationProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoContainerNameProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoSecurityPassPhraseProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoSubdomainConfigurationProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.SubDomainNotFoundException;
import org.ow2.petals.clientserverapi.topology.exception.SubdomainIsNotDynamicException;
import org.ow2.petals.clientserverapi.topology.exception.TopologyServiceException;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.jmx.api.api.PetalsJmxApiFactory;
import org.ow2.petals.jmx.api.api.exception.ConnectionErrorException;
import org.ow2.petals.jmx.api.api.exception.DuplicatedServiceException;
import org.ow2.petals.jmx.api.api.exception.MissingServiceException;
import org.ow2.petals.jmx.api.api.exception.TopologyServiceDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.TopologyServiceErrorException;
import org.ow2.petals.microkernel.api.communication.topology.TopologyService;
import org.ow2.petals.microkernel.api.communication.topology.exception.AddSubDomainException;
import org.ow2.petals.microkernel.api.communication.topology.exception.LockAlreadyAcquiredException;
import org.ow2.petals.microkernel.api.communication.topology.exception.RemoveContainerException;
import org.ow2.petals.microkernel.api.communication.topology.exception.RemoveSubDomainException;
import org.ow2.petals.microkernel.api.communication.topology.exception.TopologyException;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.SubDomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.SubDomainConfiguration.SubdomainMode;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.implementation.MutableImplementation;
import org.ow2.petals.microkernel.api.jbi.messaging.registry.EndpointRegistry;
import org.ow2.petals.microkernel.api.jbi.messaging.registry.exception.DynamicTopologyNotLockedByCurrentThreadException;
import org.ow2.petals.microkernel.api.jbi.messaging.registry.exception.DynamicTopologyNotLockedException;
import org.ow2.petals.microkernel.communication.RemoteCheckerClient;
import org.ow2.petals.microkernel.server.PetalsCompositeController;
import org.ow2.petals.topology.TopologyHelper;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.Subdomain;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.topology.generated.TransportService;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.ebmwebsourcing.easycommons.uuid.SimpleUUIDGenerator;

/**
 * The default implementation of the topology service.
 * @author Christophe Hamerling - EBM WebSourcing
 */
@Component(provides = @Interface(name = "service", signature = TopologyService.class))
public class TopologyServiceImpl implements TopologyService {

    /**
     * Pattern defining a description to put in the domain definition about the new domain of a container after
     * detachment.
     */
    private final static String DOMAIN_DESCR_DETACHMENT_PATTERN = "Domain created during container detachment from domain/subdomain '%s/%s'";

    /**
     * Pattern defining a description to put in the sub-domain definition about the new sub-domain of a container after
     * detachment.
     */
    private final static String SUBDOMAIN_DESCR_DETACHMENT_PATTERN = "Subdomain created during container detachment from domain/subdomain '%s/%s'";

    /**
     * Inner class to ping the remote container.
     * @author Roland Naudin - EBM WebSourcing
     */
    private class TopologyPinger extends TimerTask {

        @Override
        public void run() {
            for (String containerName : containerStates.keySet()) {
                if (TopologyServiceImpl.this.stopped) {
                    break;
                }
                if (!containerName.equals(TopologyServiceImpl.this.configurationService.getContainerConfiguration()
                        .getName())) {
                    if (TopologyServiceImpl.this.remoteChecker.ping(containerName)) {
                        TopologyServiceImpl.this.setContainerState(containerName,
                                ContainerState.STARTED);
                        TopologyServiceImpl.this.log.debug("Container '" + containerName
                                + "' started");
                    } else {
                        TopologyServiceImpl.this.setContainerState(containerName,
                                ContainerState.STOPPED);

                        TopologyServiceImpl.this.log.debug("Container '" + containerName
                                + "' stopped");
                    }
                }
            }
        }

    }

    /**
     * Logger wrapper.
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME));
    
    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * The service 'Registry'
     */
    @Requires(name = "registry")
    private EndpointRegistry registry;

    /**
     * The service 'PetalsCompositeController'
     */
    @Requires(name = "petalsCompositeController")
    private PetalsCompositeController petalsCompositeController;

    private ConcurrentMap<String, ContainerState> containerStates = new ConcurrentHashMap<String, ContainerState>();

    /**
     * Flag to know if the fractal component is stopping
     */
    private volatile boolean stopped;

    /**
     * The Ping service<br>
     * TODO: It must be a ping via the transporter instead of JMX
     */
    @Requires(name = "remote")
    private RemoteCheckerClient remoteChecker;

    /**
     * The timer to ping the remote containers
     */
    private Timer topologyTimer;

    /**
     * The listener to be informed of container configuration changes.
     */
    private TopologyContainerConfigurationListener containerConfigurationListener = new TopologyContainerConfigurationListener(
            this);

    /**
     * {@inheritDoc}
     */
    // TODO: Should be only a wrapper of ConfigurationService.addContainerConfiguration
    public void addContainerConfiguration(final ContainerConfiguration addedContainer)
            throws NoContainerConfigurationProvidedException, NoSubdomainConfigurationProvidedException,
            SubDomainNotFoundException, SubdomainIsNotDynamicException, AddContainerException {
        if (addedContainer == null) {
            throw new NoContainerConfigurationProvidedException();
        }

        final String subdomainName = addedContainer.getSubdomainName();
        if (subdomainName == null) {
            throw new NoSubdomainConfigurationProvidedException();
        }

        if (!this.isSubdomainDynamic(subdomainName)) {
            throw new SubdomainIsNotDynamicException(subdomainName);
        }

        try {
            this.configurationService.addContainerConfiguration(addedContainer);
            // The container states are updated through the container configuration listener
        } catch (final ConfigurationException e) {
            throw new AddContainerException(addedContainer.getName(), e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void addSubdomainConfiguration(SubDomainConfiguration addedSubdomain)
            throws AddSubDomainException {
        if (addedSubdomain == null) {
            throw new AddSubDomainException(null);
        }

        try {
            this.configurationService.addSubdomainConfiguration(addedSubdomain);
            this.updateContainerStates();
        } catch (final ConfigurationException e) {
            throw new AddSubDomainException(addedSubdomain.getName(), e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public ContainerConfiguration getContainerConfiguration(String containerName) {
        // FIXME : Default get the container configuration from the configuration service. This container information is
        // not updated at runtime!
        return this.configurationService.getContainerConfiguration(containerName);
    }

    /**
     * {@inheritDoc}
     */
    public Set<ContainerConfiguration> getContainersConfiguration(ContainerState state) {
        Set<ContainerConfiguration> result = null;

        final Set<ContainerConfiguration> set = this.configurationService
                .getContainersConfiguration();
        if (state == null) {
            result = set;
        } else {
            result = new HashSet<ContainerConfiguration>(set.size());
            for (ContainerConfiguration containerConfiguration : set) {
                if ((containerConfiguration.getState() != null)
                        && containerConfiguration.getState().equals(state)) {
                    result.add(containerConfiguration);
                }
            }
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public Set<ContainerConfiguration> getContainersConfigurationsForLocalSubdomain() {
        final Set<ContainerConfiguration> containersConfigurationForLocalSubdomain = new HashSet<ContainerConfiguration>();

        final String localSubdomainName = this.retrieveLocalSubdomainName();

        for (ContainerConfiguration containerConfiguration : this
                .retrieveAllContainersConfigurations()) {
            if (containerConfiguration.getSubdomainName().equals(localSubdomainName)) {
                containersConfigurationForLocalSubdomain.add(containerConfiguration);
            }
        }

        return containersConfigurationForLocalSubdomain;
    }

    /**
     * {@inheritDoc}
     */
    public DomainConfiguration getDomainConfiguration() {
        return this.configurationService.getDomainConfiguration();
    }

    /**
     * {@inheritDoc}
     */
    public Set<SubDomainConfiguration> getSubDomainsConfiguration() {
        return this.configurationService.getSubDomainsConfiguration();
    }

    /**
     * Check if the sub-domain identified by its given name is dynamic
     * 
     * @param subdomainName
     *            The name of the sub-domain to check. If <code>null</code> or empty, the sub-domain of the current
     *            container is used.
     * @return <code>true</code> if the sub-domain is dynamic, <code>false</code> otherwise.
     * @throws SubDomainNotFoundException
     *             The sub-domain has not been found in the current topology
     */
    private boolean isSubdomainDynamic(final String subdomainName) throws SubDomainNotFoundException {
        
        final SubDomainConfiguration subdomainConfiguration = this.configurationService.getSubDomainConfiguration(subdomainName);
        if (subdomainConfiguration != null) {
            return subdomainConfiguration.getMode().equals(SubDomainConfiguration.SubdomainMode.DYNAMIC);
        } else {
            throw new SubDomainNotFoundException(subdomainName);
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean isContainerStarted(String containerName) {
        return ContainerState.STARTED.equals(this.containerStates.get(containerName));
    }

    /**
     * {@inheritDoc}
     */
    public void removeContainerConfiguration(final ContainerConfiguration removedContainer)
            throws RemoveContainerException {
        try {
            this.configurationService.removeContainerConfiguration(removedContainer);
            this.updateContainerStates();
        } catch (final ConfigurationException e) {
            throw new RemoveContainerException(removedContainer.getName(), e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void removeSubdomainConfiguration(final SubDomainConfiguration removedSubdomain)
            throws RemoveSubDomainException {
        try {
            this.configurationService.removeSubdomainConfiguration(removedSubdomain);
            this.updateContainerStates();
        } catch (final ConfigurationException e) {
            throw new RemoveSubDomainException(removedSubdomain.getName(), e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void setContainerState(String containerName, ContainerState state) {
        this.containerStates.put(containerName, state);
    }

    @Lifecycle(step = Step.START)
    public void start() {
        this.log.start();

        this.configurationService.registerContainerConfigurationListener(this.containerConfigurationListener);

        final Set<ContainerConfiguration> containers = this.configurationService
                .getContainersConfiguration();

        for (ContainerConfiguration containerConfiguration : containers) {
            this.containerStates.put(containerConfiguration.getName(), ContainerState.UNKNOWN);
        }

        this.stopped = false;
        this.topologyTimer = new Timer("Topology Pinger");
        if (this.remoteChecker != null) {
            final ContainerConfiguration containerConfig = this.configurationService.getContainerConfiguration();
            this.topologyTimer.schedule(new TopologyPinger(), containerConfig.getTopologyPingerStartDelay(),
                    containerConfig.getTopologyPingerPeriodDelay());
        } else {
            this.log.warning("No remote checker is bound. The topology pinger is not scheduled !!");
        }

        this.log.end();
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();

        this.stopped = true;
        this.topologyTimer.cancel();

        this.configurationService.deregisterContainerConfigurationListener(this.containerConfigurationListener);

        this.log.end();
    }

    private Set<ContainerConfiguration> retrieveAllContainersConfigurations() {
        final Set<ContainerConfiguration> containersConfigurationsSet = this.configurationService
                .getContainersConfiguration();

        return containersConfigurationsSet;
    }

    private ContainerConfiguration retrieveLocalContainerConfiguration() {
        final ContainerConfiguration localContainerConfiguration = this.configurationService
                .getContainerConfiguration();

        return localContainerConfiguration;
    }

    private String retrieveLocalSubdomainName() {
        final String localContainerName = this.retrieveLocalContainerConfiguration().getName();
        String subdomainName = null;

        for (ContainerConfiguration containerConfiguration : this
                .retrieveAllContainersConfigurations()) {
            if (containerConfiguration.getName().equals(localContainerName)) {
                subdomainName = containerConfiguration.getSubdomainName();
                break;
            }
        }

        return subdomainName;
    }

    /**
     * {@inheritDoc}
     */
    public void loadTopology() throws TopologyException {
        this.log.start();
        try {
            final boolean isDynamicTopologyLocked = this.registry.lockDynamicTopology();
            if (isDynamicTopologyLocked) {
                try {
                    final Topology dynamicTopology = this.registry.getDynamicTopology();
                    if (dynamicTopology == null) {
                        // No topology previously loaded
                        this.log.info("First startup of a Petals ESB node of the topology, the dynamic topology is initialized with the topology definition.");
                        this.registry.setDynamicTopology(this.configurationService.getInitialLocalTopology());
                    } else {
                        // A dynamic topology exists, we use it
                        this.configurationService.updateTopology(dynamicTopology);
                    }
                } finally {
                    this.registry.unlockDynamicTopology();
                }
            } else {
                throw new LockAlreadyAcquiredException("Dynamic topology lock");
            }

            // Update the container state list
            this.updateContainerStates();

        } catch (final ConfigurationException e) {
            throw new TopologyException(e);
        } catch (final DynamicTopologyNotLockedException e) {
            throw new TopologyException(e);
        } catch (final DynamicTopologyNotLockedByCurrentThreadException e) {
            throw new TopologyException(e);
        } catch (final InterruptedException e) {
            throw new TopologyException(e);
        } finally {
            this.log.end();
        }
    }

    /**
     * Update the list of container states ({@link #containerStates} according to the list of containers of the current
     * configuration:
     * <ul>
     * <li>an entry is created for new container,</li>
     * <li>the entries associated to container that no more exists are removed.</li>
     * </ul>
     */
    protected void updateContainerStates() {
        final Set<ContainerConfiguration> containers = this.configurationService.getContainersConfiguration();

        for (final ContainerConfiguration containerConfiguration : containers) {
            this.containerStates.putIfAbsent(containerConfiguration.getName(), ContainerState.UNKNOWN);
        }

        final Iterator<String> itContainerStates = this.containerStates.keySet().iterator();
        while (itContainerStates.hasNext()) {
            final String containerState = itContainerStates.next();
            final Iterator<ContainerConfiguration> itContainers = containers.iterator();
            boolean isContainerExisting = false;
            while (itContainers.hasNext()) {
                final ContainerConfiguration container = itContainers.next();
                if (container.getName().equals(containerState)) {
                    isContainerExisting = true;
                    break;
                }
            }
            if (!isContainerExisting) {
                itContainerStates.remove();
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public Set<Map<String, String>> retrieveTopology(final String securityPassPhrase) throws TopologyServiceException {
        this.log.start();
        
        // Check security
        final boolean isSecurityOk = this.configurationService.isSecurityTopologyPassphraseOk(securityPassPhrase);
        if (this.log.isDebugEnabled()) {
            if (isSecurityOk) {
                this.log.debug("Security validated. Sensible information will be returned.");
            } else {
                this.log.debug("Security not validated. Sensible information will NOT be returned.");
            }
        }

        final Set<Map<String, String>> topologyMap = new HashSet<Map<String, String>>();
        topologyMap.add(this.convertDomainConfiguration(this.configurationService.getDomainConfiguration()));
        final Iterator<SubDomainConfiguration> itSubdomainsConfiguration = this.getSubDomainsConfiguration().iterator();
        while (itSubdomainsConfiguration.hasNext()) {
            final SubDomainConfiguration subdomain = itSubdomainsConfiguration.next();
            topologyMap.add(this.convertSubdomainConfiguration(subdomain, securityPassPhrase));
        }
        final Iterator<ContainerConfiguration> itContainersConfiguration = this.getContainersConfiguration(null)
                .iterator();
        while (itContainersConfiguration.hasNext()) {
            // As container state are only managed by "containerStates", we should rework container configurations
            final ContainerConfiguration reworkedContainerConfiguration = new ContainerConfiguration(
                    itContainersConfiguration.next().toMap(true));
            if (reworkedContainerConfiguration.getName().equals(
                    this.configurationService.getContainerConfiguration().getName())) {
                reworkedContainerConfiguration.setState(ContainerState.STARTED);
            } else {
                reworkedContainerConfiguration.setState(this.containerStates.get(reworkedContainerConfiguration
                        .getName()));
            }
            topologyMap.add(this.convertContainerConfiguration(reworkedContainerConfiguration, isSecurityOk));
        }
        this.log.end();
        return topologyMap;
    }

    /**
     * Convert a domain configuration into a key/value map form.
     * 
     * @param domainConfiguration
     *            The domain configuration to convert.
     * @return The domain configuration as key/value map form.
     */
    private Map<String, String> convertDomainConfiguration(final DomainConfiguration domainConfiguration) {
        this.log.start();

        try {
            final Map<String, String> domainMap = domainConfiguration.toMap();
            domainMap.put(TopologyItem.TOPOLOGY_TYPE, TopologyItem.DOMAIN);

            return domainMap;
        } finally {
            this.log.end();
        }
    }

    /**
     * Convert a sub-domain configuration into a key/value map form.
     * 
     * @param subdomainConfiguration
     *            The sub-domain configuration to convert.
     * @param securityPassPhrase
     *            Passphrase needed to return sensible information. If <code>null</code>, empty or not matching the one
     *            of the container on which the connection is established, no sensible information will be returned.
     * @return The sub-domain configuration as key/value map form.
     * @throws TopologyException
     *             An error occurs getting configuration properties
     */
    private Map<String, String> convertSubdomainConfiguration(final SubDomainConfiguration subdomainConfiguration,
            final String securityPassPhrase) throws TopologyException {
        this.log.start();

        try {
            Map<String, String> subdomainMap = new HashMap<String, String>();
            subdomainMap.put(TopologyItem.TOPOLOGY_TYPE, TopologyItem.SUBDOMAIN);
            subdomainMap.put(SubdomainPropertyName.CONF_SUBDOMAIN_NAME, subdomainConfiguration.getName());
            if (subdomainConfiguration.getDescription() != null) {
                subdomainMap.put(SubdomainPropertyName.CONF_SUBDOMAIN_DESCRIPTION,
                        subdomainConfiguration.getDescription());
            }

            if (SubDomainConfiguration.SubdomainMode.DYNAMIC.equals(subdomainConfiguration.getMode())) {
                subdomainMap.put(SubdomainPropertyName.CONF_SUBDOMAIN_MODE, SubdomainPropertyValues.MODE_DYNAMIC);
            } else {
                subdomainMap.put(SubdomainPropertyName.CONF_SUBDOMAIN_MODE, SubdomainPropertyValues.MODE_STATIC);
            }
            subdomainMap.put(SubdomainPropertyName.CONF_NETWORK_TIME_SYNCHRONIZED,
                    Boolean.toString(subdomainConfiguration.isNetworkTimeSynchronized()));

            if (subdomainConfiguration.getRegistryImplementation() != null) {
                subdomainMap.put(SubdomainPropertyName.CONF_REGISTRY_IMPLEMENTATION,
                        subdomainConfiguration.getRegistryImplementation());
            } else {
                subdomainMap.put(SubdomainPropertyName.CONF_REGISTRY_IMPLEMENTATION, "");
            }

            try {
                final Properties registryProperties = this.petalsCompositeController
                        .getRegistryConfiguration(securityPassPhrase);
                for (final String propertyName : registryProperties.stringPropertyNames()) {
                    subdomainMap.put(propertyName, registryProperties.getProperty(propertyName));
                }
                return subdomainMap;
            } catch (final PetalsCompositeCtrlException e) {
                throw new TopologyException(e);
            }
        } finally {
            this.log.end();
        }
    }

    /**
     * Convert a container configuration into a key/value map form.
     * 
     * @param containerConfiguration
     *            The container configuration to convert.
     * @param isSecurityOk
     *            if <code>true</code>, sensible information are returned. Otherwise, sensible information are skipped.
     * @return The container configuration as key/value map form.
     */
    private Map<String, String> convertContainerConfiguration(final ContainerConfiguration containerConfiguration,
            final boolean isSecurityOk) {
        this.log.start();

        try {
            final Map<String, String> containerMap = containerConfiguration.toMap(isSecurityOk);
            containerMap.put(TopologyItem.TOPOLOGY_TYPE, TopologyItem.CONTAINER);

            return containerMap;
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void detachContainer() throws SubdomainIsNotDynamicException, DetachContainerException {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());

        final ContainerConfiguration currentContainerConfig = this.configurationService.getContainerConfiguration();
        final SubDomainConfiguration currentSubdomainConfig = this.configurationService.getSubDomainConfiguration();
        final DomainConfiguration currentDomainConfig = this.configurationService.getDomainConfiguration();

        try {
            this.checkIfDetachmentIsPossible();
            // Stop the messaging sub-system
            this.petalsCompositeController.stopMessagingSubsystem();
            try {
                // Unregister the local endpoints from the registry without undeploying them
                this.registry.removeAllLocalEndpoints();

                // Free all JBI artifacts
                this.petalsCompositeController.freeAllJbiArtifacts();

                // Remove the current container from all existing containers of the sub-domain.
                // We can do it using JMX API invoking 'TopologyService.removeContainerConfiguration' on each container
                // of the sub-domain. But how to manage use-cases where the remote container is unreachable ?
                // A solution is to use a mechanism based on event listeners:
                // - Each container register a listener on the dynamic topology
                // - When an event 'modify' occurs, the container will check differences and process them.
                this.log.debug("Updating dynamic topology in registry ...");
                final Topology topology = this.registry.getDynamicTopology();
                final Subdomain currentSubdomain = TopologyHelper.findSubdomain(currentSubdomainConfig.getName(),
                        topology);
                final Iterator<Container> itContainers = currentSubdomain.getContainer().iterator();
                while (itContainers.hasNext()) {
                    if (currentContainerConfig.getName().equals(itContainers.next().getName())) {
                        itContainers.remove();
                    }
                }

                // The new target topology is written in the registry
                try {
                    if (this.registry.lockDynamicTopology()) {
                        try {
                            this.registry.setDynamicTopology(topology);
                            this.log.debug("Dynamic topology updated in registry.");
                        } finally {
                            try {
                                this.registry.unlockDynamicTopology();
                            } catch (final RegistryException e) {
                                this.log.warning("An error occurs releasing the dynamic topology lock", e);
                            }
                        }
                    } else {
                        throw new DetachContainerException("Unable to acquire the dynamic topology lock.");
                    }
                } catch (final InterruptedException e) {
                    this.log.error("Interrupted exception during trying to lock the dynamic topology", e);
                    throw new DetachContainerException(e);
                }

                // We can stop the registry, before reconfiguring the container and restart it
                this.log.debug("Stopping registry ...");
                this.petalsCompositeController.stopRegistryComponent();
                this.log.debug("Registry stopped");

                // A stand-alone topology is persisted in the working area to be able to restart the
                // container in its own topology. The topology is persisted before to be stored in the
                // registry, in this way, if an error occurs writing in the registry, a restart is
                // sufficient to restore state and to not join the initial topology.
                // The stand-alone topology is created re-using some configuration information:
                // - the container name to avoid to change the content of the server.properties
                // - the container hostname that has not to be changed
                // - the JMX configuration
                // - the port of the fast remote transporter
                // - the configuration of the web-service service
                // - the registry configuration is re-used, and as the domain name and the sub-domain name has been
                // changed, the registries (the registry leaved, and the new registry created) are distinct.
                final Topology standaloneTopology = new Topology();
                final Domain standaloneDomain = new Domain();
                standaloneTopology.setDomain(standaloneDomain);
                final SimpleUUIDGenerator uuidGenerator = new SimpleUUIDGenerator();
                final String newDomainName = uuidGenerator.getNewID();
                standaloneDomain.setName(newDomainName);
                final Subdomain standaloneSubdomain = new Subdomain();
                standaloneDomain.getSubDomain().add(standaloneSubdomain);
                standaloneSubdomain.setMode(org.ow2.petals.topology.generated.SubdomainMode.DYNAMIC);
                final String newSubdomainName = uuidGenerator.getNewID();
                standaloneSubdomain.setName(newSubdomainName);
                standaloneSubdomain.setRegistryImplementation(currentSubdomainConfig.getRegistryImplementation());
                final Container standaloneContainer = new Container();
                standaloneSubdomain.getContainer().add(standaloneContainer);
                standaloneSubdomain.getAny().addAll(currentSubdomainConfig.getExtraParameters());
                standaloneContainer.setHost(currentContainerConfig.getHost());
                standaloneContainer.setName(currentContainerConfig.getName());
                standaloneContainer.setUser(currentContainerConfig.getUser());
                standaloneContainer.setPassword(currentContainerConfig.getPassword());
                final JmxService standaloneJmxService = new JmxService();
                standaloneJmxService.setRmiPort(currentContainerConfig.getJmxRMIConnectorPort());
                standaloneContainer.setJmxService(standaloneJmxService);
                final TransportService standaloneTransportService = new TransportService();
                standaloneTransportService.setTcpPort(currentContainerConfig.getTCPPort());
                standaloneContainer.setTransportService(standaloneTransportService);
                this.configurationService.persistTopology(standaloneTopology);

                // Deregister all other containers member of the initial sub-domain
                for (final ContainerConfiguration containerToRemove : this.configurationService
                        .getContainersConfiguration()) {
                    if (containerToRemove.getSubdomainName().equals(currentSubdomain.getName())
                            && !containerToRemove.getName().equals(currentContainerConfig.getName())) {
                        this.removeContainerConfiguration(containerToRemove);
                    }
                }

                this.log.debug("Updating container configuration ...");
                final DomainConfiguration newDomainCfg = new DomainConfiguration();
                newDomainCfg.setName(newDomainName);
                newDomainCfg.setDescription(String.format(DOMAIN_DESCR_DETACHMENT_PATTERN,
                        currentDomainConfig.getName(), currentSubdomainConfig.getName()));
                this.configurationService.reconfigureDomain(newDomainCfg);
                final SubDomainConfiguration newSubdomainCfg = new SubDomainConfiguration();
                newSubdomainCfg.setName(newSubdomainName);
                newSubdomainCfg.setDescription(String.format(SUBDOMAIN_DESCR_DETACHMENT_PATTERN,
                        currentDomainConfig.getName(), currentSubdomainConfig.getName()));
                newSubdomainCfg.setMode(SubdomainMode.DYNAMIC);
                newSubdomainCfg.setRegistryImplementation(currentSubdomainConfig.getRegistryImplementation());
                newSubdomainCfg.setExtraParameters(currentSubdomainConfig.getExtraParameters());
                newSubdomainCfg.setNetworkTimeSynchronized(currentSubdomainConfig.isNetworkTimeSynchronized());
                this.configurationService.reconfigureSubdomain(newSubdomainCfg);
                this.configurationService.reconfigureContainer(newSubdomainCfg);
                this.log.debug("Container configuration updated");

                this.log.debug("Re-Starting registry ...");
                this.petalsCompositeController.startRegistryComponent();
                this.log.debug("Registry re-started.");

                // We load the new stand-alone topology of the detached container (registering the stand-alone topology
                // as dynamic topology into the registry)
                this.loadTopology();

                this.petalsCompositeController.startMessagingSubsystem();

                // Endpoints are redeployed into the registry of the joined sub-domain using the recovery process,
                // and JBI artifacts are recreated using the recovering service, as a new start of the container
                this.log.debug("Recovering system ...");
                this.petalsCompositeController.recoverSystem();
                this.log.debug("System recovered.");

            } catch (final RegistryException e) {
                throw new DetachContainerException(e);
            } catch (final ConfigurationException e) {
                throw new DetachContainerException(e);
            } finally {
                this.petalsCompositeController.startMessagingSubsystem();
            }

        } catch (final SubdomainIsNotDynamicException e) {
            throw e;
        } catch (final PetalsCompositeCtrlException e) {
            this.log.warning("An error occurs at PetalsCompositeController level", e);
            throw new DetachContainerException(e);
        } catch (final PetalsException e) {
            this.log.warning("An error occurs attaching the container to another sub-domain", e);
            throw new DetachContainerException(e);
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void attachContainerTo(final ContainerConfiguration container, final String securityPassPhrase)
            throws NoContainerConfigurationProvidedException, NoContainerNameProvidedException,
            NoSubdomainConfigurationProvidedException, SubdomainIsNotDynamicException, InconsistencyException,
            NoSecurityPassPhraseProvidedException, AttachContainerException {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
        try {

            if (container == null) {
                throw new NoContainerConfigurationProvidedException();
            }

            if (securityPassPhrase == null || securityPassPhrase.isEmpty()) {
                throw new NoSecurityPassPhraseProvidedException();
            }

            final String containerName = container.getName();
            if (containerName == null) {
                throw new NoContainerNameProvidedException();
            }
            final String subdomainName = container.getSubdomainName();
            if (subdomainName == null) {
                throw new NoSubdomainConfigurationProvidedException();
            }
            try {
                this.log.info("Attaching the current container to the subdomain '" + subdomainName
                        + "' using the container '" + containerName + "'");
                // Create a JMX connection to the container of the sub-domain in which the current container must be
                // attached.
                // The service 'JMXService' cannot be used because the container of the sub-domain to join is not a
                // member of the initial topology of the current container.
                final JMXClient petalsJmxClient = PetalsJmxApiFactory.getInstance().createJMXClient(
                        container.getHost(),
                        container.getJmxRMIConnectorPort() == 0 ? JMXClient.DEFAULT_PORT : container
                                .getJmxRMIConnectorPort(), container.getUser(), container.getPassword());
                try {
                    final SubDomainConfiguration targetSubdomainCfg = new SubDomainConfiguration();
                    final DomainConfiguration targetDomainCfg = new DomainConfiguration();
                    this.checkIfAttachmentIsPossible(petalsJmxClient, container, securityPassPhrase, targetDomainCfg,
                            targetSubdomainCfg);
                    this.petalsCompositeController.stopMessagingSubsystem();
                    try {

                        // Unregister the local endpoints from the registry without undeploying them
                        this.registry.removeAllLocalEndpoints();

                        // Free all JBI artifacts
                        this.petalsCompositeController.freeAllJbiArtifacts();

                        // Stop the registry, before reconfiguring the container and restart it
                        this.log.debug("Stopping registry ...");
                        this.petalsCompositeController.stopRegistryComponent();
                        this.log.debug("Registry stopped");
                        try {
                            this.log.debug("Updating container configuration ...");
                            this.configurationService.reconfigureDomain(targetDomainCfg);
                            this.configurationService.reconfigureSubdomain(targetSubdomainCfg);
                            this.configurationService.reconfigureContainer(targetSubdomainCfg);

                            // Register containers member of the new sub-domain
                            final Set<Map<String, String>> retrievedTopology = petalsJmxClient
                                    .getTopologyServiceClient().retrieveTopology(securityPassPhrase);
                            for (final Map<String, String> retrievedTopologyMap : retrievedTopology) {
                                if (TopologyItem.CONTAINER.equals(retrievedTopologyMap.get(TopologyItem.TOPOLOGY_TYPE))) {
                                    final ContainerConfiguration containerConfig = new ContainerConfiguration(
                                            retrievedTopologyMap);
                                    this.addContainerConfiguration(containerConfig);
                                    // The container states are updated through the container configuration listener
                                }
                            }

                            // Add the current new container to all existing containers of the target sub-domain.
                            // We can do it using JMX API invoking 'TopologyService.addContainerConfiguration' on
                            // each container of the target sub-domain. But how to manage use-cases where the remote
                            // container is unreachable ?
                            // A solution is to use a mechanism based on event listeners:
                            // - Each container register a listener on the target dynamic topology
                            // - When an event 'modify' occurs, the container will check differences and process
                            // them.
                            // NOTE: This mechanism requires a running registry !!! That's why it is implemented
                            // just after to have restarted the registry.

                            this.log.debug("Container configuration updated");

                        } finally {
                            this.log.debug("Re-Starting registry ...");
                            this.petalsCompositeController.startRegistryComponent();
                            this.log.debug("Registry re-started.");

                            // Update the dynamic topology in the registry to inform all other containers about the
                            // new current container.
                            // See comments above for more information.
                            this.log.debug("Updating dynamic topology in registry ...");
                            final Topology topology = this.registry.getDynamicTopology();
                            final Subdomain targetSubdomain = TopologyHelper.findSubdomain(subdomainName, topology);

                            final Container containerAttached = new Container();
                            final ContainerConfiguration currentContainerConfig = this.configurationService
                                    .getContainerConfiguration();
                            containerAttached.setName(currentContainerConfig.getName());
                            containerAttached.setDescription(currentContainerConfig.getDescription());
                            containerAttached.setHost(currentContainerConfig.getHost());
                            containerAttached.setUser(currentContainerConfig.getUser());
                            containerAttached.setPassword(currentContainerConfig.getPassword());
                            final JmxService jmxService = new JmxService();
                            jmxService.setRmiPort(currentContainerConfig.getJmxRMIConnectorPort());
                            containerAttached.setJmxService(jmxService);
                            final TransportService transportService = new TransportService();
                            transportService.setTcpPort(currentContainerConfig.getTCPPort());
                            containerAttached.setTransportService(transportService);

                            targetSubdomain.getContainer().add(containerAttached);

                            // The new target topology is persisted in the working area to be able to restart the
                            // container in the right topology. The topology is persisted before to be stored in the
                            // registry, in this way, if an error occurs writing in the registry, a restart is
                            // sufficient to restore state and to join the target topology.
                            this.configurationService.persistTopology(topology);

                            // The new target topology is written in the registry
                            try {
                                if (this.registry.lockDynamicTopology()) {
                                    try {
                                        this.registry.setDynamicTopology(topology);
                                        this.log.debug("Dynamic topology updated in registry.");
                                    } finally {
                                        try {
                                            this.registry.unlockDynamicTopology();
                                        } catch (final RegistryException e) {
                                            this.log.warning("An error occurs releasing the dynamic topology lock", e);
                                        }
                                    }
                                } else {
                                    throw new AttachContainerException(subdomainName, containerName,
                                            "Unable to acquire the dynamic topology lock.");
                                }
                            } catch (final InterruptedException e) {
                                this.log.error("Interrupted exception during trying to lock the dynamic topology", e);
                                throw new AttachContainerException(subdomainName, containerName, e);
                            }
                        }

                    } catch (final RegistryException e) {
                        throw new AttachContainerException(subdomainName, containerName, e);
                    } finally {
                        this.petalsCompositeController.startMessagingSubsystem();
                    }

                    // Endpoints are redeployed into the registry of the joined sub-domain using the recovery process,
                    // and JBI artifacts are recreated using the recovering service, as a new start of the container
                    this.log.debug("Recovering system ...");
                    this.petalsCompositeController.recoverSystem();
                    this.log.debug("System recovered.");

                    // We notify the Petals composite controller that the configuration has changed to adjust
                    // states of Fractal components (for example, the Embedded Registry must be stopped if the
                    // container is not running in a stand-alone topology).
                    this.petalsCompositeController.onConfigurationUpdated();

                    this.log.info("Current container attached to the subdomain '" + subdomainName + "'");

                } finally {
                    petalsJmxClient.disconnect();
                }

            } catch (final ConnectionErrorException e) {
                throw new AttachContainerException(subdomainName, containerName, e);
            } catch (final DuplicatedServiceException e) {
                throw new AttachContainerException(subdomainName, containerName, e);
            } catch (final MissingServiceException e) {
                throw new AttachContainerException(subdomainName, containerName, e);
            } catch (final TopologyServiceDoesNotExistException e) {
                throw new AttachContainerException(subdomainName, containerName, e);
            } catch (final TopologyServiceErrorException e) {
                throw new AttachContainerException(subdomainName, containerName, e);
            } catch (final SubdomainIsNotDynamicException e) {
                throw e;
            } catch (final InconsistencyException e) {
                throw e;
            } catch (final PetalsCompositeCtrlException e) {
                this.log.warning("An error occurs at PetalsCompositeController level", e);
                throw new AttachContainerException(subdomainName, containerName, e);
            } catch (final PetalsException e) {
                this.log.warning("An error occurs attaching the container to another sub-domain", e);
                throw new AttachContainerException(subdomainName, containerName, e);
            }
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * <p>
     * Check that the container attachment can be realized. During checks, the target sub-domain configuration is
     * retrieved. Checks applied are:
     * <ul>
     * <li>the sub-domain of the current container is 'dynamic'</li>
     * <li>the given container exists in the target topology</li>
     * <li>the sub-domain of the given container exists in the target topology</li>
     * <li>the sub-domain of the given container is the one in which it is declared in the target topology</li>
     * <li>the target sub-domain mode is 'dynamic'</li>
     * <li>the registry implementation of the target sub-domain must be the same than the one of the current sub-domain</li>
     * </ul>
     * </p>
     * 
     * @param petalsJmxClient
     *            A JMX client to the given target container
     * @param container
     *            The given container member of the target sub-domain. Not <code>null</code>, and all required fields
     *            have been previously checked
     * @param securityPassPhrase
     *            The security pass-phrase of the given container member of the target sub-domain
     * @param targetDomainConf
     *            The target domain configuration fill during checks
     * @param targetSubdomainConf
     *            The target sub-domain configuration fill during checks
     */
    private void checkIfAttachmentIsPossible(final JMXClient petalsJmxClient, final ContainerConfiguration container,
            final String securityPassPhrase, final DomainConfiguration targetDomainConf,
            final SubDomainConfiguration targetSubdomainConf)
            throws TopologyServiceErrorException, TopologyServiceDoesNotExistException, ConnectionErrorException,
            InconsistencyException, PetalsException {

        // check if the sub-domain of the current container is dynamic
        if (!this.isSubdomainDynamic(null)) {
            throw new SubdomainIsNotDynamicException(null);
        }

        final String containerName = container.getName();
        final String subdomainName = container.getSubdomainName();

        final Set<Map<String, String>> targetTopologyMaps = petalsJmxClient.getTopologyServiceClient()
                .retrieveTopology(securityPassPhrase);
        boolean containerFound = false;
        final Properties subdomainProperties = new Properties();
        for (final Map<String, String> targetTopologyMap : targetTopologyMaps) {
            final String topologyType = targetTopologyMap.get(TopologyService.TopologyItem.TOPOLOGY_TYPE);
            if (topologyType.equals(TopologyService.TopologyItem.CONTAINER)) {
                if (containerName.equals(targetTopologyMap
                        .get(TopologyService.ContainerPropertyName.CONF_CONTAINER_NAME))) {
                    containerFound = true;
                    targetSubdomainConf.setName(targetTopologyMap
                            .get(TopologyService.ContainerPropertyName.CONF_SUBDOMAIN_NAME));
                }
            } else if (topologyType.equals(TopologyService.TopologyItem.SUBDOMAIN)) {
                subdomainProperties.putAll(targetTopologyMap);
            } else if (topologyType.equals(TopologyService.TopologyItem.DOMAIN)) {
                targetDomainConf.setName(targetTopologyMap.get(TopologyService.DomainPropertyName.CONF_DOMAIN_NAME));
                targetDomainConf.setDescription(targetTopologyMap
                        .get(TopologyService.DomainPropertyName.CONF_DOMAIN_DESCRIPTION));
            }
        }
        if (!containerFound) {
            throw new InconsistentContainerNameException(containerName);
        }
        if (!subdomainName.equals(targetSubdomainConf.getName())) {
            throw new InconsistentSubdomainNameException(targetSubdomainConf.getName(), subdomainName);
        }
        if (!this.findTargetSubdomain(targetTopologyMaps, subdomainName, targetSubdomainConf)) {
            // This exception has not to occur because a container of the target topology is a member of the expected
            // domain. So if this error occurs, an inconsistency exists on the remote container about sub-domain that
            // should be due to a bug in configuration loading of the remote container.
            throw new InconsistencySubdomainException("Unable to retrieve the target sub-domain in the target topology");
        }
        if (targetSubdomainConf.getMode().equals(SubdomainMode.STATIC)) {
            throw new SubdomainIsNotDynamicException(subdomainName);
        }

        // Check the registry implementation consistency
        final SubDomainConfiguration currentSubdomainConfiguration = this.configurationService
                .getSubDomainConfiguration();
        final String currentRegistryImpl = currentSubdomainConfiguration.getRegistryImplementation();
        final String targetRegistryImpl = targetSubdomainConf.getRegistryImplementation();
        if ((currentRegistryImpl == null && targetRegistryImpl != null)
                || (currentRegistryImpl != null && targetRegistryImpl == null)
                || (currentRegistryImpl != null && targetRegistryImpl != null && !currentRegistryImpl
                        .equals(targetRegistryImpl))) {
            throw new InconsistentSubdomainRegistryImplException(currentRegistryImpl, targetRegistryImpl);
        }

        // Check that the registry implementation is in the classloader
        final MutableImplementation registryImplInstance;
        try {
            // As SubDomainConfiguration.getRegistryImplementation() returns null for default implementation loaded from
            // classloader, we must use the JMX API to get the registry implementation really used by the sub-domain to
            // join
            final Class<?> registryImpl = Class.forName(petalsJmxClient.getTopologyServiceClient()
                    .retrieveRegistryImplementation());
            registryImplInstance = (MutableImplementation) registryImpl.newInstance();
        } catch (final ClassNotFoundException e) {
            throw new AttachContainerException(subdomainName, containerName, "Registry implementation '"
                    + currentRegistryImpl + "' not found");
        } catch (final InstantiationException e) {
            throw new AttachContainerException(subdomainName, containerName, e);
        } catch (final IllegalAccessException e) {
            throw new AttachContainerException(subdomainName, containerName, e);
        }

        // Set the registry implementation configuration
        final List<Object> extraParameters = new ArrayList<Object>();
        extraParameters.add(registryImplInstance.getTopologyPartConfiguration(subdomainProperties));
        targetSubdomainConf.setExtraParameters(extraParameters);

        // The container IP address cannot be 'localhost'. If possible, it is replaced by its real IP address
        final String containerHost = this.configurationService.getContainerConfiguration().getHost();
        try {
            if (InetAddress.getByName(containerHost).isLoopbackAddress()) {
                try {
                    final String newIpAddress = InetAddress.getLocalHost().getHostAddress();
                    this.configurationService.getContainerConfiguration().setHost(newIpAddress);
                    this.log.config("New container host IP address: " + newIpAddress);
                } catch (final UnknownHostException e) {
                    throw new InconsistencyException(
                            "Unable to replace the loopback address of the container by its real address");
                }
            }
        } catch (final UnknownHostException e) {
            this.log.warning("Unable to determine if the container IP address (" + containerHost + ")is the loopback",
                    e);
        }
    }

    /**
     * <p>
     * Check that the container detachment can be realized. Checks applied are:
     * <ul>
     * <li>the sub-domain of the current container is 'dynamic'.</li>
     * </ul>
     * </p>
     * 
     * @throws SubdomainIsNotDynamicException
     *             The current sub-domain is not <b>dynamic</b>.
     */
    private void checkIfDetachmentIsPossible() throws SubdomainIsNotDynamicException, TopologyServiceException {

        // check if the sub-domain of the current container is dynamic
        if (!this.isSubdomainDynamic(null)) {
            throw new SubdomainIsNotDynamicException(null);
        }
    }

    /**
     * Retrieve the target sub-domain configuration from the target topology.
     * 
     * @param targetTopology
     *            The target topology which the target sub-domain will be searched, as {@link Set} of {@link Map}
     * @param targetSubdomainName
     *            The target sub-domain name to search
     * @param targetSubdomainConf
     *            The target sub-domain configuration to initialize
     * @return <code>true</code> if the target sub-domain has been found in the target topology, <code>false</code>
     *         otherwise.
     */
    private boolean findTargetSubdomain(final Set<Map<String, String>> targetTopology,
            final String targetSubdomainName, final SubDomainConfiguration targetSubdomainConf) {
        boolean subdomainFound = false;
        for (final Map<String, String> targetTopologyMap : targetTopology) {
            final String topologyType = targetTopologyMap.get(TopologyService.TopologyItem.TOPOLOGY_TYPE);
            if (topologyType.equals(TopologyService.TopologyItem.SUBDOMAIN)) {
                if (targetSubdomainName.equals(targetTopologyMap
                        .get(TopologyService.SubdomainPropertyName.CONF_SUBDOMAIN_NAME))) {
                    subdomainFound = true;
                    final String mode = targetTopologyMap
                            .get(TopologyService.SubdomainPropertyName.CONF_SUBDOMAIN_MODE);
                    if ("static".equals(mode)) {
                        targetSubdomainConf.setMode(SubdomainMode.STATIC);
                    } else {
                        targetSubdomainConf.setMode(SubdomainMode.DYNAMIC);
                    }
                    final String targetRegistryImpl = targetTopologyMap
                            .get(TopologyService.SubdomainPropertyName.CONF_REGISTRY_IMPLEMENTATION);
                    targetSubdomainConf.setRegistryImplementation(targetRegistryImpl != null
                            && targetRegistryImpl.isEmpty() ? null : targetRegistryImpl);
                }
            }
        }
        return subdomainFound;
    }

    /**
     * {@inheritDoc}
     */
    public String retrieveRegistryImplementation() throws PetalsException {

        return this.petalsCompositeController.getRegistryImplementation();
    }
}
