/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.container.lifecycle;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.Bootstrap;
import javax.jbi.component.Component;
import javax.jbi.management.ComponentLifeCycleMBean;
import javax.jbi.management.LifeCycleMBean;
import javax.jbi.messaging.MessagingException;
import javax.management.ObjectName;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.extensions.Controller;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.jbi.descriptor.original.generated.ClassLoaderDelegationType;
import org.ow2.petals.jbi.descriptor.original.generated.Component.SharedLibrary;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.container.ComponentContextCommunication;
import org.ow2.petals.microkernel.api.container.ComponentLifeCycle;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.container.exception.ContainerServiceException;
import org.ow2.petals.microkernel.api.container.thread.BootstrapThread;
import org.ow2.petals.microkernel.api.container.thread.ComponentLifeCycleThread;
import org.ow2.petals.microkernel.api.jbi.component.PetalsComponentContext;
import org.ow2.petals.microkernel.api.jbi.management.AdminService;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.messaging.PetalsDeliveryChannel;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.jbi.messaging.RoutingException;
import org.ow2.petals.microkernel.api.jbi.messaging.registry.EndpointRegistry;
import org.ow2.petals.microkernel.api.jbi.servicedesc.JBIServiceEndpointImpl;
import org.ow2.petals.microkernel.api.jbi.servicedesc.PetalsServiceEndpoint;
import org.ow2.petals.microkernel.api.jbi.servicedesc.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.microkernel.api.server.FractalHelper;
import org.ow2.petals.microkernel.api.system.SystemStateService;
import org.ow2.petals.microkernel.api.system.repository.RepositoryService;
import org.ow2.petals.microkernel.communication.jndi.client.JNDIService;
import org.ow2.petals.microkernel.jbi.component.context.ComponentContextImpl;
import org.ow2.petals.microkernel.jbi.component.context.ComponentInitialContext;
import org.ow2.petals.microkernel.jbi.component.context.InstallationContextImpl;
import org.ow2.petals.microkernel.server.exception.InvalidCompBasedArchiException;
import org.ow2.petals.microkernel.system.classloader.ClassLoaderService;
import org.ow2.petals.microkernel.system.classloader.PetalsClassLoader;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * @author Adrien Louis - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 * @author wjoseph - EBM WebSourcing
 */
@org.objectweb.fractal.fraclet.annotations.Component(provides = @Interface(name = Installer.FRACTAL_SRV_ITF_NAME, signature = Installer.class))
public class InstallerImpl implements Installer {

    /**
     * Bootstrap classloader id prefix.
     */
    public static final String BS_PREFIX = "BS";

    /**
     * Component classloader id prefix.
     */
    public static final String COMP_PREFIX = "COMP";

    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(Installer.COMPONENT_LOGGER_NAME));

    /**
     * JBI-Management :: Admin Service
     */
    @Requires(name = "admin")
    private AdminService adminService;

    /**
     * Platform :: Loader Service
     */
    @Requires(name = "classloader")
    private ClassLoaderService classloaderService;

    /**
     * JBI-Messaging :: Router Service
     */
    @Requires(name = "router")
    private RouterService router;

    /**
     * Platform :: Repository Service
     */
    @Requires(name = "repository")
    private RepositoryService repositoryService;

    /**
     * JBI-Messaging :: EndpointRegistry
     */
    @Requires(name = "endpoint")
    private EndpointRegistry endpointRegistry;

    /**
     * Communication :: JNDIService
     */
    @Requires(name = "jndi")
    private JNDIService jndiService;

    /**
     * Container :: Container Service
     */
    @Requires(name = "container")
    private ContainerService container;

    /**
     * Platform component :: SystemState Service
     */
    @Requires(name = "systemstate")
    private SystemStateService recoverySrv;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * The Fractal component
     */
    @Controller
    private org.objectweb.fractal.api.Component fractalComponent;

    /**
     * The container configuration
     */
    private ContainerConfiguration containerConfiguration;

    private Context context;

    private String state;

    private Jbi jbiDescriptor;

    private URL installationRootURL;

    private String installationRootPath;

    private String workingRootPath;

    /**
     * The consumer endpoint
     */
    private PetalsServiceEndpoint address;

    private InitialContext initialContext;

    private InstallationContextImpl installContext;

    private PetalsClassLoader bootstrapClassLoader;

    private BootstrapThread bootstrapThread;

    private Object activitySynchronizer;

    private PetalsComponentContext componentContext;

    private org.ow2.petals.jbi.descriptor.original.generated.Component componentDescription;

    private String componentName;

    private Bootstrap jbiBootstrap;

    /**
     * Create the Installer that will manage installation / uninstallation of
     * the JBI component. The related bootstrap is created and initialize here.
     */
    public InstallerImpl() {
        super();
        this.state = UNINSTALLED;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.container.lifecycle.InstallerInterface#init(java.util.Map)
     */
    public void init(final Context ctxt) throws IOException, JBIException, NamingException {

        this.context = ctxt;

        this.state = UNINSTALLED;

        this.activitySynchronizer = new Object();

        this.jbiDescriptor = this.context.getDescriptor();

        this.componentDescription = this.jbiDescriptor.getComponent();

        this.componentName = this.componentDescription.getIdentification().getName();

        try {
            this.installationRootPath = new File(this.context.getInstallUrl().toURI()).getPath();
        } catch (URISyntaxException e) {
            throw new IOException("Installation root path is not valid", e);
        }
        this.installationRootURL = this.context.getInstallUrl();

        this.workingRootPath = this.repositoryService.getComponentWorkDirectory(this.componentName)
                .getAbsolutePath();

        this.address = new JBIServiceEndpointImpl();
        this.address.setType(EndpointType.CONSUMER);
        this.address.getLocation().setComponentName(this.componentName);
        this.address.getLocation().setContainerName(this.containerConfiguration.getName());
        this.address.getLocation().setSubdomainName(this.containerConfiguration.getSubdomainName());

        this.initialContext = this.createComponentInitialContext();

        // 1. create the component context
        try {
            this.componentContext = this.createComponentContext();
        } catch (final InvalidCompBasedArchiException e) {
            throw new JBIException(new InvalidCompBasedArchiException(e));
        }

        // 2. create the bootstrap classloader
        this.bootstrapClassLoader = this.createBootstrapClassloader();

        // 3. create the bootstrap object
        this.jbiBootstrap = this.loadBootstrap(this.bootstrapClassLoader);

        // 5. create installation context
        this.installContext = new InstallationContextImpl(this.componentDescription,
                this.componentContext, true);

        // 6. init the Bootstrap thread that execute actions on the Bootstrap
        this.bootstrapThread = this.createBootstrapThread(this.bootstrapClassLoader,
                this.jbiBootstrap);

        // 7. init bootstrap
        this.initBootstrap(true);

    }

    public void shutdown() throws JBIException {
        this.bootstrapThread.cleanUp();

        // unload bootstrap classloader
        this.classloaderService.deleteClassLoader(InstallerImpl.BS_PREFIX + this.componentName);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.InstallerMBean#getInstallerConfigurationMBean()
     */
    public ObjectName getInstallerConfigurationMBean() throws JBIException {
        return this.bootstrapThread.getExtensionMBeanName();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.InstallerMBean#getInstallRoot()
     */
    public String getInstallRoot() {
        return this.componentContext.getInstallRoot();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.InstallerMBean#install()
     */
    public ObjectName install() throws JBIException {
        this.log.start();
        ObjectName componentMBeanName = null;

        try {
            if (UNINSTALLED.equals(this.state)) {
                synchronized (this.activitySynchronizer) {
                    // init the boostrap
                    try {
                        this.bootstrapThread.onInstall();
                    } finally {
                        this.bootstrapThread.cleanUp();
                    }

                    // create the component life cycle
                    componentMBeanName = this.createComponentLifeCycleMBean();
                    this.setState(INSTALLED);

                }
            } else {
                throw new JBIException("The Installer can not be installed in this state: "
                        + this.state);
            }
        } catch (JBIException e) {
            this.log.error(e.getMessage(), e);
            throw e;
        }

        this.log.info("Component '" + this.componentName + "' installed");

        this.log.end();

        return componentMBeanName;
    }

    /**
     * Uninstall the JBI component. Call the Bootstrap onUninstall() and
     * cleanup() methods, unregister the jbiComponent object,
     * 
     * @throws JBIException
     *             the component is already uninstalled
     */
    public void uninstall() throws JBIException {
        this.log.start();

        try {
            if (this.adminService.getServiceUnitsLifeCyclesForComponent(this.componentName).size() != 0) {
                throw new JBIException(
                        "Can not uninstall a component which holds deployed Service Unit(s)");
            }

            final ComponentLifeCycleMBean componentLifeCycle = this.container
                    .getComponentLifecycleByName(this.componentName);
            if (componentLifeCycle == null) {
                throw new JBIException("The component '" + this.componentName
                        + "' can not be uninstalled as it is not installed!");
            }

            String componentState = componentLifeCycle.getCurrentState();
            if ((INSTALLED.equals(this.state))
                    && (LifeCycleMBean.SHUTDOWN.equals(componentState) || LifeCycleMBean.UNKNOWN
                            .equals(componentState))) {
                synchronized (this.activitySynchronizer) {
                    this.doUninstall();
                    this.setState(UNINSTALLED);
                }
            } else {
                throw new JBIException("The component '" + this.componentName
                        + "' can not be uninstalled in the Installer state: " + this.state
                        + " and Component state: " + componentState);
            }
        } catch (JBIException e) {
            this.log.error(e.getMessage(), e);
            throw e;
        }

        this.log.info("Component '" + this.componentName + "' uninstalled");

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.InstallerMBean#isInstalled()
     */
    public boolean isInstalled() {
        return INSTALLED.equals(this.state);
    }

    /**
     * {@inheritDoc}
     */
    public String getComponentName() {
        return this.componentName;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getAddress()
     */
    public PetalsServiceEndpoint getAddress() {
        return this.address;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getEndpointRegistry()
     */
    public EndpointRegistry getEndpointRegistry() {
        return this.endpointRegistry;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getInitialContext()
     */
    public InitialContext getInitialContext() {
        return this.initialContext;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getInstallationRoot()
     */
    public String getInstallationRoot() {
        return this.installationRootPath;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getJBIDescriptor()
     */
    public Jbi getJBIDescriptor() {
        return this.jbiDescriptor;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getAdminService()
     */
    public AdminService getAdminService() {
        return this.adminService;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication# getContainerService()
     */
    public ContainerService getContainerService() {
        return this.container;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication# getRouterService()
     */
    public RouterService getRouterService() {
        return this.router;
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication# getWorkspaceRoot()
     */
    public String getWorkspaceRoot() {
        return this.workingRootPath;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.container.lifecycle.ComponentContextCommunication#getLogger
     * ()
     */
    public Logger getLogger() {
        return this.log.getLogger();
    }

    /**
     * {@inheritDoc}
     */
    public Component getComponent() {
        return this.componentContext.getComponent();
    }

    /**
     * {@inheritDoc}
     */
    public void initBootstrap(boolean install) throws JBIException {
        this.installContext.setInstall(install);
        try {
            this.bootstrapThread.init(this.installContext);
        } catch (JBIException e) {
            this.bootstrapThread.cleanUp();
            throw e;
        }

    }

    /**
     * Start of the petals component
     * 
     * @throws PetalsException
     * @throws JBIException
     * 
     */
    @Lifecycle(step = Step.START)
    public void start() throws JBIException {
        this.log.start();

        this.containerConfiguration = this.configurationService.getContainerConfiguration();

        this.log.end();
    }

    /**
     * Stop of the petals component
     * 
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();

        try {
            this.bootstrapThread.shutdownThread();
        } catch (JBIException e) {
            this.log.warning("Failed to stop the Bootstrap thread : " + e.getMessage());
        }

        this.log.end();
    }

    /**
     * Set the <code>State</code> reference of the current state. An
     * <code>LifeCycleStateEventAbstract</code> is thrown.
     * 
     * @param state
     * @throws JBIException
     */
    private void setState(String state) throws JBIException {
        this.state = state;

        try {
            this.recoverySrv.updateComponentInstallationState(this.getComponentName(), state);
        } catch (Exception e) {
            throw new JBIException("Installer state can't be persisted for recovery", e);
        }
    }

    /**
     * <p>
     * Create the classLoader for Bootstrap
     * </p>
     * <p>
     * The bootstrap classloader depends on share library classloader to be able to check configuration parameter as
     * JDBC Driver, JMS Driver provided through SL.
     * </p>
     * 
     * @return The bootstrap classloader.
     */
    private PetalsClassLoader createBootstrapClassloader() {
        this.log.call();

        final List<String> classpath = this.componentDescription.getBootstrapClassPath().getPathElement();

        // Shared libraries used by this component
        final List<SharedLibrary> slList = this.componentDescription.getSharedLibraryList();
        final List<String> slNameList = new ArrayList<String>();
        if (slList != null) {
            for (final SharedLibrary sl : slList) {
                slNameList.add(sl.getContent().trim() + "-" + sl.getVersion());
            }
        }

        final boolean parentFirst = !ClassLoaderDelegationType.SELF_FIRST
                .equals(this.componentDescription.getBootstrapClassLoaderDelegation());

        return this.createClassLoader(BS_PREFIX, classpath, parentFirst, slNameList);
    }

    /**
     * Create a classloader with the specified parameters.
     * 
     * @param prefix
     *            bootstrap or component prefix
     * @param classpath
     *            classpath of the bootstrap/component
     * @param parentFirst
     *            delegation mode
     * @param slNameList
     *            list of the name of the required sharedLibraries
     * @return a classloader, or <code>null</code> if an error occurs.
     */
    private PetalsClassLoader createClassLoader(final String prefix, final List<String> classpath,
            final boolean parentFirst, final List<String> slNameList) {
        this.log.start();

        // prepare bootstrap and component classloader
        final URL[] baseUrls = { this.installationRootURL };

        try {
            return (PetalsClassLoader) this.classloaderService.createComponentClassLoader(
                    prefix + this.componentName, baseUrls, classpath, parentFirst, slNameList);
        } catch (final PetalsException pe) {
            // just log when some error occur
            this.log.error("Failed to create bootstrap class loader" + pe.getMessage(), pe);
            return null;
        } finally {
            this.log.end();
        }
    }

    /**
     * Create the classloader for the component. Be careful, we use the
     * classpath set in the InstallationContext. (during the installation phase,
     * the Bootstrap can modify the component's classpath by setting it in the
     * InstallationContext object.)
     * 
     * @return classloader for the component
     */
    private PetalsClassLoader createComponentClassloader() {
        this.log.start();

        final List<String> classpath = this.installContext.getClassPathElements();

        // Delegation model
        final boolean parentFirst = !ClassLoaderDelegationType.SELF_FIRST
                .equals(this.componentDescription.getComponentClassLoaderDelegation());

        // Shared libraries used by this component
        final List<SharedLibrary> slList = this.componentDescription.getSharedLibraryList();
        final List<String> slNameList = new ArrayList<String>();
        if (slList != null) {
            for (final SharedLibrary sl : slList) {
                slNameList.add(sl.getContent().trim() + "-" + sl.getVersion());
            }
        }

        final PetalsClassLoader componentClassloader = this.createClassLoader(COMP_PREFIX, classpath,
                parentFirst, slNameList);

        this.log.end();
        return componentClassloader;
    }

    /**
     * Create the thread that will be used to process component installation
     * steps. The thread-classloader is set with the given paremeter.
     * 
     * @param bootstrap
     *            classloader
     * @param bootstrap
     *            instance
     * @return the BootstrapThread
     */
    private BootstrapThread createBootstrapThread(PetalsClassLoader classloader, Bootstrap bootstrap) {
        BootstrapThread it = new BootstrapThread(this.getComponentName(), bootstrap,
                this.containerConfiguration.getTaskTimeout(), this.log);
        it.setContextClassLoader(classloader);
        it.start();

        return it;
    }

    /**
     * Create the thread that will be used to process component life cycle
     * steps. The thread-classloader is set with the given paremeter.
     * 
     * @param classloader
     * @return the BootstrapThread
     */
    private ComponentLifeCycleThread createComponentLifeCycleThread(Component component,
            PetalsClassLoader classloader) {
        ComponentLifeCycleThread clt = new ComponentLifeCycleThread(component,
                this.containerConfiguration.getTaskTimeout(), this.log);
        clt.setContextClassLoader(classloader);
        clt.start();
        return clt;
    }

    /**
     * Register the installer engine/binding in the life cycle manager service
     * 
     * @return
     * @throws JBIException
     * @throws PetalsException
     */
    private ObjectName createComponentLifeCycleMBean() throws JBIException {
        this.log.start();

        // 1. create component classloader
        PetalsClassLoader cl = this.createComponentClassloader();

        // 2. create the component object
        Component jbiComponent = this.loadComponent(cl);

        // 3. set the jbiComponent reference to the contextImpl
        this.componentContext.setComponent(jbiComponent);

        // 4. create the componentLifeCycle thread
        ComponentLifeCycleThread lifeCycleThread = this.createComponentLifeCycleThread(
                jbiComponent, cl);

        try {
            // 5. Creation of the life cycle
            final ComponentLifeCycle componentLifecycle = this.container
                    .createComponentLifeCycle(this.componentDescription);

            // 6. Start the life cycle
            final String name = this.componentDescription.getIdentification().getName();
            this.container.startComponentLifeCycle(name);

            // Initialization of the life cycle
            componentLifecycle.init(this.componentContext, lifeCycleThread, this.context);

            final ObjectName objectName = this.container.registerComponentLifeCycleMBean(name);

            // 7. inform the router about the component installation
            this.router.addComponent(this.componentContext);

            return objectName;
        } catch (RoutingException e) {
            throw new JBIException(e);
        } catch (PetalsException e) {
            throw new JBIException(e);
        } finally {
            this.log.end();
        }
    }

    /**
     * Create and return the component context
     * 
     * @return the component context
     * @throws InvalidCompBasedArchiException
     */
    private PetalsComponentContext createComponentContext() throws InvalidCompBasedArchiException {

        try {
            final org.objectweb.fractal.api.Component containerComposite = FractalHelper
                    .getParentComponent(this.fractalComponent);
            final org.objectweb.fractal.api.Component jbiMngtComposite = FractalHelper
                    .getParentComponent(containerComposite);

            final PetalsComponentContext componentContextImpl = new ComponentContextImpl(
                    (ComponentContextCommunication) this.fractalComponent.getFcInterface("/content"), jbiMngtComposite);
            this.context.setComponentContext(componentContextImpl);
            return componentContextImpl;
        } catch (final NoSuchInterfaceException | PetalsException e) {
            throw new InvalidCompBasedArchiException(e);
        }
    }

    /**
     * The initial JNDI context for all component. If this initial context is
     * null, it will be created
     * 
     * @return the initial JNDI context for all component. If this initial
     *         context is null, it will be created
     * @throws PetalsException
     *             if {@link InitialContext} can't be created
     */
    private InitialContext createComponentInitialContext() throws NamingException {
        ComponentInitialContext componentInitialContext = null;
        componentInitialContext = new ComponentInitialContext(this.jndiService.getUsersContext());

        return componentInitialContext;
    }

    /**
     * load the bootstrap, with its own classloader
     * 
     * @return
     */
    private Bootstrap loadBootstrap(PetalsClassLoader classloader) throws JBIException {
        this.log.call();

        Class<?> bootstrapClass;
        Bootstrap jbiBootstrapTmp = null;
        try {
            bootstrapClass = classloader.loadClass(this.componentDescription
                    .getBootstrapClassName().trim());

            // if it is null there is a bug somewhere: loadClass should throw an exception
            assert bootstrapClass != null;
            
            jbiBootstrapTmp = (Bootstrap) bootstrapClass.newInstance();

        } catch (final ClassNotFoundException e) {
            throw new JBIException("Class not found: "
                    + this.componentDescription.getBootstrapClassName(), e);
        } catch (final InstantiationException e) {
            throw new JBIException("Could not instantiate : "
                    + this.componentDescription.getBootstrapClassName(), e);
        } catch (final IllegalAccessException e) {
            throw new JBIException("Illegal access on: "
                    + this.componentDescription.getBootstrapClassName(), e);
        } catch (final SecurityException e) {
            // this can happen because we use URLClassLoaders
            throw new JBIException("Security exception on: "
                    + this.componentDescription.getComponentClassName(), e);
        }
        return jbiBootstrapTmp;
    }

    /**
     * Load the component, within its own classloader
     * 
     * @return the component
     */
    private Component loadComponent(PetalsClassLoader classloader) throws JBIException {
        this.log.call();
        Component jbiComponent;

        Class<?> componentClass;
        try {
            componentClass = classloader.loadClass(this.componentDescription
                    .getComponentClassName().getContent().trim());

            // if it is null there is a bug somewhere: loadClass should throw an exception
            assert componentClass != null;

            jbiComponent = (Component) componentClass.newInstance();

        } catch (final ClassNotFoundException e) {
            throw new JBIException("Class not found: "
                    + this.componentDescription.getComponentClassName(), e);
        } catch (final InstantiationException e) {
            throw new JBIException("Could not instantiate : "
                    + this.componentDescription.getComponentClassName(), e);
        } catch (final IllegalAccessException e) {
            throw new JBIException("Illegal access on: "
                    + this.componentDescription.getComponentClassName(), e);
        } catch (final SecurityException e) {
            // this can happen because we use URLClassLoaders
            throw new JBIException("Security exception on: "
                    + this.componentDescription.getComponentClassName(), e);
        }

        return jbiComponent;
    }

    /**
     * Do the bootstrap uninstallation. All the exceptions occurred are logged.<br />
     * If errors occurs, only the first exception will be thrown.
     * 
     * @throws JBIException
     *             : the first error generated
     */
    private void doUninstall() {
        this.log.start();

        // Ensure that the DeliveryChannel of the component is closed
        // (if the component is in an unknown state, for instance)
        try {
            final PetalsDeliveryChannel channel = this.componentContext.getDeliveryChannelInstance();
            if ((channel != null) && channel.isOpened()) {
                this.log.debug("close the component DeliveryChannel");
                channel.close();
            }
        } catch (MessagingException e) {
            this.log.error("Unable to close the delivery channel: ", e);
        }

        // Ensure that all endpoints of the component are unregistered
        try {
            this.componentContext.deregisterAllEndpoints();
        } catch (final JBIException e) {
            this.log.error("Unable to unregistered all remaining endpoints: ", e);
        }

        // call uninstall on bootstrap
        try {
            this.bootstrapThread.onUninstall();
        } catch (JBIException e) {
            this.log.error("Unable to uninstall the bootstap: ", e);
        }

        // clean the bootstrap thread
        try {
            this.bootstrapThread.cleanUp();
        } catch (JBIException e) {
            this.log.error("Unable to clean up the bootstap: ", e);
        }

        // unregister the component life-cycle mbean
        try {
            this.container.unregisterComponentLifeCycleMBean(this.componentDescription.getIdentification().getName());
        } catch (final ContainerServiceException e) {
            this.log.error("Unable to unregister the component lifecycle: ", e);
        }
        // unregister the optional extension mbean
        try {
            if (this.componentContext.getComponent().getLifeCycle() != null) {
                ObjectName extName = this.componentContext.getComponent().getLifeCycle()
                        .getExtensionMBeanName();
                if (extName != null) {
                    this.adminService.getMBeanServer().unregisterMBean(extName);
                }
            }
        } catch (Exception e) {
            this.log.error("Unable to unregister the extension of the component: ", e);
        }
        // inform the router about the component uninstallation
        try {
            this.router.removeComponent(this.componentContext);
        } catch (RoutingException e) {
            this.log.error("Unable to inform the router about the component uninstallation: ", e);
        }

        // remove the fractal component
        try {
            this.container.removeComponentLifeCycle(this.componentDescription.getIdentification().getName());
        } catch (PetalsException e) {
            this.log.error("Unable to remove the fractal component: ", e);
        }

        // clean the component context
        this.componentContext.clear();

        // unload component classloader
        this.classloaderService.deleteClassLoader(InstallerImpl.COMP_PREFIX + this.componentName);

        this.log.end();
    }

}
