/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.exchange;

import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.URI;
import java.sql.SQLException;
import java.util.Map;
import java.util.Set;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.jbi.messaging.exchange.MessageExchangeWrapper;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.microkernel.api.jbi.messaging.exchange.PersistedPetalsMessageExchange;
import org.ow2.petals.microkernel.api.jbi.servicedesc.PetalsServiceEndpoint;
import org.ow2.petals.microkernel.api.system.persistence.PersistenceService;

/**
 *
 * This implementation of {@link PetalsMessageExchange} (and {@link PersistedPetalsMessageExchange}) wraps a
 * {@link PetalsMessageExchange} and can persists it.
 * 
 * This has nothing to do with {@link MessageExchangeWrapper} that handles owner checking for the components.
 * 
 * @author Nicolas Oddoux - EBM WebSourcing
 * @author vnoel
 * 
 */
public class PersistedMessageExchangeImpl implements PersistedPetalsMessageExchange {

    private static final long serialVersionUID = 481103387216052125L;

    /**
     * We use exchange properties to store this information to avoid adding extra class members to the generic
     * implementation of {@link MessageExchange} ({@link PetalsMessageExchange}) that is actually stored/monitored.
     */
    private final static String PROPERTY_MONITORED = "org.ow2.petals.microkernel.api.jbi.messaging.exchange.PersistedPetalsMessageExchange.isMonitored";

    private final static String PROPERTY_PERSISTED = "org.ow2.petals.microkernel.api.jbi.messaging.exchange.PersistedPetalsMessageExchange.isPersisted";
    
    /**
     * It is not {@code null} when the message has been persisted.
     */
    private transient PersistenceService persistenceService;

    /**
     * This can be set to null when the message is persisted. But when it is set back, it must always conform to the
     * value of {@link #exchangeId}
     */
    private PetalsMessageExchange messageExchange;

    private final String exchangeId;

    public PersistedMessageExchangeImpl(final PetalsMessageExchange messageExchange) {
        this.exchangeId = messageExchange.getExchangeId();
        this.messageExchange = messageExchange;
    }

    @Override
    public Fault createFault() throws MessagingException {
        this.checkExchange();

        return this.messageExchange.createFault();
    }
    
    @Override
    public NormalizedMessage createMessage() throws MessagingException {
        this.checkExchange();

        return this.messageExchange.createMessage();
    }

    @Override
    public PetalsServiceEndpoint getConsumerEndpoint() {
        this.checkExchange();

        return (PetalsServiceEndpoint) this.messageExchange.getConsumerEndpoint();
    }

    @Override
    public PetalsServiceEndpoint getEndpoint() {
        this.checkExchange();

        return (PetalsServiceEndpoint) this.messageExchange.getEndpoint();
    }

    @Override
    public Exception getError() {
        this.checkExchange();

        return this.messageExchange.getError();
    }
    
    @Override
    public Fault getFault() {
        this.checkExchange();

        return this.messageExchange.getFault();
    }

    @Override
    public QName getInterfaceName() {
        this.checkExchange();

        return this.messageExchange.getInterfaceName();
    }

    @Override
    public NormalizedMessage getMessage(String name) {
        this.checkExchange();

        return this.messageExchange.getMessage(name);
    }

    @Override
    public Map<String, NormalizedMessage> getMessages() {
        this.checkExchange();

        return this.messageExchange.getMessages();
    }

    @Override
    public QName getOperation() {
        this.checkExchange();

        return this.messageExchange.getOperation();
    }

    @Override
    public URI getPattern() {
        this.checkExchange();

        return this.messageExchange.getPattern();
    }

    @Override
    public Object getProperty(String name) {
        this.checkExchange();

        return this.messageExchange.getProperty(name);
    }

    @Override
    public Set<?> getPropertyNames() {
        this.checkExchange();

        return this.messageExchange.getPropertyNames();
    }

    @Override
    public Role getRole() {
        this.checkExchange();

        return this.messageExchange.getRole();
    }

    @Override
    public QName getService() {
        this.checkExchange();

        return this.messageExchange.getService();
    }

    @Override
    public ExchangeStatus getStatus() {
        this.checkExchange();

        return this.messageExchange.getStatus();
    }
    
    @Override
    public String getTermination() {
        return this.messageExchange.getTermination();
    }
    
    @Override
    public boolean isTransacted() {
        this.checkExchange();

        return this.messageExchange.isTransacted();
    }

    @Override
    public void setEndpoint(ServiceEndpoint endpoint) {
        this.checkExchange();

        this.messageExchange.setEndpoint(endpoint);
    }

    @Override
    public void setError(Exception error) {
        this.checkExchange();

        this.messageExchange.setError(error);
    }

    @Override
    public void setFault(Fault fault) throws MessagingException {
        this.checkExchange();

        this.messageExchange.setFault(fault);
    }

    @Override
    public void setInterfaceName(QName interfaceName) {
        this.checkExchange();

        this.messageExchange.setInterfaceName(interfaceName);
    }

    @Override
    public void setMessage(NormalizedMessage msg, String name) throws MessagingException {
        this.checkExchange();

        this.messageExchange.setMessage(msg, name);
    }
    
    @Override
    public void setOperation(QName name) {
        this.checkExchange();

        this.messageExchange.setOperation(name);
    }

    @Override
    public void setProperty(String name, Object obj) {
        this.checkExchange();

        this.messageExchange.setProperty(name, obj);
    }

    @Override
    public void setRole(Role role) {
        this.checkExchange();

        this.messageExchange.setRole(role);
    }

    @Override
    public void setService(QName service) {
        this.checkExchange();

        this.messageExchange.setService(service);
    }

    @Override
    public void setStatus(ExchangeStatus status) throws MessagingException {
        this.checkExchange();

        this.messageExchange.setStatus(status);
    }
    
    @Override
    public void setTermination(String termination) {
        this.checkExchange();

        this.messageExchange.setTermination(termination);
    }

    /**
     * Persist the wrapped exchange.
     * 
     * Hold the PersistenceService reference to retrieve the exchange from the persistence service when necessary.
     */
    @Override
    public void persistExchange(PersistenceService persistenceService) throws SQLException,
            IOException {
        if (isPersisted() || isMonitored()) {
            persistenceService.updateMessageExchange(this.messageExchange, true, null);
        } else {
            persistenceService.addMessageExchange(this.messageExchange, true, null);
        }
        this.messageExchange = null;
        this.persistenceService = persistenceService;
    }
    
    @Override
    public boolean isMessageExchangeStored() {
        return this.messageExchange == null;
    }

    // TODO remove monitoring?
    @Override
    public void setMonitored(final boolean monitored) {
        setProperty(PROPERTY_MONITORED, monitored);
    }

    @Override
    public boolean isMonitored() {
        final Object monitored = getProperty(PROPERTY_MONITORED);
        return monitored != null && monitored instanceof Boolean && ((boolean) monitored);
    }

    // TODO remove persistence?
    @Override
    public void setPersisted(final boolean persisted) {
        setProperty(PROPERTY_PERSISTED, persisted);
    }

    @Override
    public boolean isPersisted() {
        final Object persisted = getProperty(PROPERTY_PERSISTED);
        return persisted != null && persisted instanceof Boolean && ((boolean) persisted);
    }

    @Override
    public void setConsumerEndpoint(ServiceEndpoint serviceEndpoint) {
        checkExchange();

        this.messageExchange.setConsumerEndpoint(serviceEndpoint);
    }

    @Override
    public void setPattern(URI pattern) {
        checkExchange();

        this.messageExchange.setPattern(pattern);
    }

    @Override
    public void setDoChecks(boolean doChecks) {
        checkExchange();

        this.messageExchange.setDoChecks(doChecks);
    }

    @Override
    public String getExchangeId() {
        return exchangeId;
    }

    /**
     * Check if the exchange is in memory. if not, retrieve it from the
     * persistence manager
     * 
     * @throws SQLException
     */
    private void checkExchange() {
        if (this.messageExchange == null) {
            try {
                this.messageExchange = this.persistenceService.getMessageExchange(exchangeId);
                this.persistenceService = null;
            } catch (SQLException e) {
                throw new RuntimeException("Failed to retrieve Message Exchange with id '"
                        + exchangeId + "' from the persistence service", e);
            } catch (IOException e) {
                throw new RuntimeException("Failed to retrieve Message Exchange with id '"
                        + exchangeId + "' from the persistence service", e);
            } catch (ClassNotFoundException e) {
                throw new RuntimeException("Failed to retrieve Message Exchange with id '"
                        + exchangeId + "' from the persistence service", e);
            }
        }
    }

    private void writeObject(final ObjectOutputStream s) throws IOException {
        // ensure the exchange is in memory before serializing
        checkExchange();

        s.defaultWriteObject();
    }
}
