/**
 * Copyright (c) 2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.transport;

import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange.Role;

import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.transport.util.TransportSendContext;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * 
 * @author Victor Noel - Linagora
 *
 */
public abstract class AbstractTransporterImpl implements Transporter {

    /**
     * the Transporter listener (router) Fractal component
     */
    @Requires(name = "transportlistener")
    protected TransportListener transportListener;

    protected final LoggingUtil log;

    /**
     * Map of exchanges that are blocked during a synchronous send
     * 
     * The key is the message exchange id, and the value is the local (in case of distributed implementation) wrapper
     */
    private ConcurrentMap<String, PetalsMessageExchange> pendingSyncExchanges;

    /**
     * Traffic flag to handle stop of traffic
     * 
     * It is volatile to ensure that every thread has its real value
     */
    private volatile boolean stopTraffic;

    protected AbstractTransporterImpl(final LoggingUtil log) {
        this.log = log;
    }

    /**
     * To be called to deliver (by the transport listener) the exchange to the right endpoint
     */
    protected void deliverExchange(final PetalsMessageExchange exchange) {
        this.log.start();
        if (this.log.isDebugEnabled()) {
            this.log.debug(exchange.getExchangeId()
                    + ": we notify the transport listener of this new exchange, it will be dispatched to the right endpoint");
        }
        this.transportListener.onExchange(exchange);
        this.log.end();
    }

    /**
     * To be called to notify the transport listener that the exchange was sent
     */
    protected void notifySent(final PetalsMessageExchange exchange) {
        this.log.start();
        if (this.log.isDebugEnabled()) {
            this.log.debug(exchange.getExchangeId() + ": we notify the transport listener that the exchange was sent");
        }
        this.transportListener.exchangeSent(exchange);
        this.log.end();
    }

    /**
     * This is meant to be called BEFORE the implementation of
     * {@link #sendSync(PetalsMessageExchange, TransportSendContext)}.
     */
    protected void preSendSync(final PetalsMessageExchange exchange, final TransportSendContext transportContext)
            throws TransportException {

        if (!exchange.getStatus().equals(ExchangeStatus.ACTIVE)) {
            throw new TransportException("Synchronous send is forbidden when sending an acknowledgement");
        }

        this.checkTransporterState();
    }

    /**
     * This is meant to be called BEFORE the implementation of
     * {@link #send(PetalsMessageExchange, TransportSendContext)}.
     */
    protected void preSend(final PetalsMessageExchange exchange, final TransportSendContext transportContext)
            throws TransportException {
        this.checkTransporterState();
    }

    /**
     * @return true if it timed out!
     * @throws InterruptedException
     */
    protected boolean waitForAnswerOrTimeout(final PetalsMessageExchange exchange, final Role preWaitRole)
            throws TransportException, InterruptedException {
        this.log.start();

        // wait for a response or a timeout reached: interruption will just throw an exception
        final Long deadline = (Long) exchange.getProperty(RouterService.PROPERTY_ROUTER_TIMEOUT_DEADLINE);
        if (this.log.isDebugEnabled()) {
            this.log.debug("Let's wait for an answer, "
                    + (deadline == null ? "no timeout" : "timeout left is " + (deadline - System.currentTimeMillis()))
                    + " and before send, role was " + preWaitRole + "...");
        }
        // the loop protects from spurious wakeup
        while (true) {
            // we can start to wait directly, because of the synchronize, there is no way that the notify happened
            // before we called wait!
            try {
                if (deadline != null) {
                    final long timeout = deadline - System.currentTimeMillis();
                    // if timeout is <= 0, then we just don't want to wait (wait(0) means no timeout!)
                    if (timeout > 0) {
                        exchange.wait(timeout);
                    }
                } else {
                    exchange.wait();
                }
            } catch (final InterruptedException e) {
                this.log.debug("The wait was interrupted...");
                throw e;
            }

            this.log.debug("The wait was released so let's check the condition...");

            // we must check with the one in sendSync because depending on the transport the exchange could have been
            // exchanged (as with NIO) or not (as with local).
            PetalsMessageExchange newExchange = getFromSendSync(exchange.getExchangeId());
            // if it's null, it's because we are in local mode and the exchange was removed
            if (newExchange == null) {
                // so we can just use the same exchange as before
                newExchange = exchange;
            }
            final Role newRole = newExchange.getRole();
            if (!preWaitRole.equals(newRole)) {
                // we got an answer if the exchange was changed from the other role to the new one
                if (this.log.isDebugEnabled()) {
                    this.log.debug("... the condition was met, current role is " + newRole + ", we can stop waiting!");
                }
                this.log.end();
                return false;
            } else if (deadline != null) {
                this.log.debug("... the condition wasn't met, let's check timeout...");
                // we check for timeout ONLY if we haven't got the answer since what matters is to get the answer
                if (deadline < System.currentTimeMillis()) {
                    this.log.debug("... and yes, we timed out!");
                    this.log.end();
                    return true;
                } else {
                    // else we continue to wait
                    this.log.debug("... none of the conditions we were waiting for happened: it must have been the famous mysterious spurious wake-up!");
                }
            } else {
                this.log.debug(
                        "... the condition wasn't met and there is no timeout: it must have been the famous mysterious spurious wake-up!");
            }
        }
    }

    @Override
    public void stopTraffic() {
        this.log.start();

        this.stopTraffic = true;

        // Awake all the pending 'SYNC_SEND' issuer threads to abort the invocation
        // Note: after stopTraffic has been set to true, noone can add anything to pendingSyncExchanges
        // but we are not completely sure because there can't be an interleaving of threads so there is this loop
        while (!this.pendingSyncExchanges.isEmpty()) {
            final Iterator<PetalsMessageExchange> iterator = this.pendingSyncExchanges.values().iterator();
            while (iterator.hasNext()) {
                final PetalsMessageExchange exchange = iterator.next();
                synchronized (exchange) {
                    exchange.notifyAll();
                }
                iterator.remove();
            }
        }

        this.log.end();
    }


    protected void storeForSendSync(final PetalsMessageExchange exchange) {
        final PetalsMessageExchange removed = this.pendingSyncExchanges.put(exchange.getExchangeId(), exchange);
        this.log.assertOrLog(removed == null || removed == exchange,
                "A different exchange with the same ID was already in processing, it shouldn't happen");
    }

    protected void replaceInSendSync(final PetalsMessageExchange exchange) {
        final PetalsMessageExchange removed = this.pendingSyncExchanges.put(exchange.getExchangeId(), exchange);
        this.log.assertOrLog(removed != null,
                "A different exchange with the same ID was already in processing, it shouldn't happen");
    }

    protected PetalsMessageExchange removeOfSendSync(final String exchangeId) {
        return this.pendingSyncExchanges.remove(exchangeId);
    }

    protected PetalsMessageExchange getFromSendSync(final String exchangeId) {
        return this.pendingSyncExchanges.get(exchangeId);
    }

    /**
     * Check if the transporter is running, if not, raise an exception.
     * 
     * @throws TransportException
     */
    protected void checkTransporterState() throws TransportException {
        if (this.stopTraffic) {
            Thread.currentThread().interrupt();
            throw new TransportException("The Transporter traffic is stopped");
        }
    }

    protected boolean asBooleanOrFalse(final Object o) {
        if (o instanceof Boolean) {
            return (Boolean) o;
        } else {
            return false;
        }
    }

    @Lifecycle(step = Step.START)
    public void start() {
        this.log.start();

        // TODO this queue should be sizable (what does it means?! in the configuration file??)
        this.pendingSyncExchanges = new ConcurrentHashMap<String, PetalsMessageExchange>(100);

        this.log.end();
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();
    }
}
