/**
 * Copyright (c) 2014-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.topology;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.clientserverapi.topology.exception.AddContainerException;
import org.ow2.petals.clientserverapi.topology.exception.NoContainerConfigurationProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoSubdomainConfigurationProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.SubDomainNotFoundException;
import org.ow2.petals.clientserverapi.topology.exception.SubdomainIsNotDynamicException;
import org.ow2.petals.microkernel.api.communication.topology.TopologyService;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.MissingServerPropertyException;
import org.ow2.petals.microkernel.api.configuration.exception.NoServerPropertiesProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.NoTopologyProvidedException;
import org.ow2.petals.microkernel.configuration.ConfigurationServiceImpl;
import org.ow2.petals.microkernel.configuration.ConfigurationTestUtils;
import org.ow2.petals.microkernel.configuration.LocalServerPropertiesBuilder;
import org.ow2.petals.topology.generated.SubdomainMode;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * Unit tests of {@link TopologyServiceImpl#addContainerConfiguration(ContainerConfiguration)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class TopologyServiceImpl_AddContainerConfigurationTest extends AbstractTopologyServiceImplTest {

    private final static String CONTAINER_NAME_1 = "sample-1";

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    /**
     * <p>
     * Check {@link TopologyServiceImpl#addContainerConfiguration(ContainerConfiguration)} against:
     * <ul>
     * <li>no container configuration is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the missing container configuration
     * </p>
     */
    @Test(expected = NoContainerConfigurationProvidedException.class)
    public void testNoContainerConfiguration() throws InvocationTargetException, NoTopologyProvidedException,
            NoServerPropertiesProvidedException, MissingServerPropertyException, ConfigurationException,
            NoContainerConfigurationProvidedException, NoSubdomainConfigurationProvidedException,
            SubdomainIsNotDynamicException, AddContainerException, SubDomainNotFoundException, IOException {

        // Create a basic sample topology
        final Topology basicSampleTopology = ConfigurationTestUtils.createBasicSampleTopology();

        // Create and initialize the configuration service
        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(serverProps, basicSampleTopology);

            final TopologyService topoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(topoSvc, "configurationService", cfgSvc);
            this.startTopologyService(topoSvc);

            try {
                topoSvc.addContainerConfiguration(null);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#addContainerConfiguration(ContainerConfiguration)} against:
     * <ul>
     * <li>no sub-domain in which the container must be added is defined</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the missing sub-domain information
     * </p>
     */
    @Test(expected = NoSubdomainConfigurationProvidedException.class)
    public void testNoSubdomain() throws InvocationTargetException, NoTopologyProvidedException,
            NoServerPropertiesProvidedException, MissingServerPropertyException, ConfigurationException,
            NoContainerConfigurationProvidedException, NoSubdomainConfigurationProvidedException,
            SubdomainIsNotDynamicException, AddContainerException, SubDomainNotFoundException, IOException {

        // Create a basic sample topology
        final Topology basicSampleTopology = ConfigurationTestUtils.createBasicSampleTopology();

        // Create and initialize the configuration service
        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(serverProps, basicSampleTopology);

            final TopologyService topoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(topoSvc, "configurationService", cfgSvc);
            this.startTopologyService(topoSvc);

            try {
                final ContainerConfiguration newContainerConfiguration = new ContainerConfiguration();

                topoSvc.addContainerConfiguration(newContainerConfiguration);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#addContainerConfiguration(ContainerConfiguration)} against:
     * <ul>
     * <li>the sub-domain in which the container must be added does not exist</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the sub-domain not found
     * </p>
     */
    @Test(expected = SubDomainNotFoundException.class)
    public void testSubdomainNotFound() throws InvocationTargetException, NoTopologyProvidedException,
            NoServerPropertiesProvidedException, MissingServerPropertyException, ConfigurationException,
            NoContainerConfigurationProvidedException, NoSubdomainConfigurationProvidedException,
            SubdomainIsNotDynamicException, AddContainerException, SubDomainNotFoundException, IOException {

        // Create a basic sample topology
        final Topology basicSampleTopology = ConfigurationTestUtils.createBasicSampleTopology();

        // Create and initialize the configuration service
        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(serverProps, basicSampleTopology);

            final TopologyService topoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(topoSvc, "configurationService", cfgSvc);
            this.startTopologyService(topoSvc);

            try {
                final ContainerConfiguration newContainerConfiguration = new ContainerConfiguration();
                newContainerConfiguration.setSubdomainName("subdomain-name-not-found");

                topoSvc.addContainerConfiguration(newContainerConfiguration);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#addContainerConfiguration(ContainerConfiguration)} against:
     * <ul>
     * <li>the sub-domain in which the container must be added is static</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs because the sub-domain mode is not <b>dynamic</b>
     * </p>
     */
    @Test(expected = SubdomainIsNotDynamicException.class)
    public void testSubdomainNotDynamic() throws InvocationTargetException, NoTopologyProvidedException,
            NoServerPropertiesProvidedException, MissingServerPropertyException, ConfigurationException,
            NoContainerConfigurationProvidedException, NoSubdomainConfigurationProvidedException,
            SubdomainIsNotDynamicException, AddContainerException, SubDomainNotFoundException, IOException {

        // Create a basic sample topology
        final Topology basicSampleTopology = ConfigurationTestUtils.createBasicSampleTopology();
        basicSampleTopology.getDomain().getSubDomain().get(0).setMode(SubdomainMode.STATIC);

        // Create and initialize the configuration service
        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(serverProps, basicSampleTopology);

            final TopologyService topoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(topoSvc, "configurationService", cfgSvc);
            this.startTopologyService(topoSvc);

            try {
                final ContainerConfiguration newContainerConfiguration = new ContainerConfiguration();
                newContainerConfiguration.setSubdomainName(ConfigurationTestUtils.SUBDOMAIN_NAME_0);

                topoSvc.addContainerConfiguration(newContainerConfiguration);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#addContainerConfiguration(ContainerConfiguration)} against:
     * <ul>
     * <li>the container to add is valid,</li>
     * <li>the sub-domain in which the container must be added is valid.</li>
     * <li>the container is added again</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>On the first add:
     * <ul>
     * <li>No error occurs</li>
     * <li>The container is correctly added</li>
     * <li>The topology contains 2 containers</li>
     * </ul>
     * </li>
     * <li>An error occurs on the second add because the container already exist</li>
     * </ul>
     * </p>
     */
    @Test
    public void testContainerProvidedIsValid() throws InvocationTargetException, NoTopologyProvidedException,
            NoServerPropertiesProvidedException, MissingServerPropertyException, ConfigurationException,
            NoContainerConfigurationProvidedException, NoSubdomainConfigurationProvidedException,
            SubdomainIsNotDynamicException, AddContainerException, SubDomainNotFoundException, IOException {

        // Create a basic sample topology
        final Topology basicSampleTopology = ConfigurationTestUtils.createBasicSampleTopology();
        basicSampleTopology.getDomain().getSubDomain().get(0).setMode(SubdomainMode.DYNAMIC);

        // Create and initialize the configuration service
        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(serverProps, basicSampleTopology);

            final TopologyService topoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(topoSvc, "configurationService", cfgSvc);
            this.startTopologyService(topoSvc);

            try {
                final ContainerConfiguration newContainerConfiguration = new ContainerConfiguration();
                newContainerConfiguration.setName(CONTAINER_NAME_1);
                newContainerConfiguration.setSubdomainName(ConfigurationTestUtils.SUBDOMAIN_NAME_0);

                topoSvc.addContainerConfiguration(newContainerConfiguration);

                assertNotNull("container 0 no more exists",
                        cfgSvc.getContainerConfiguration(ConfigurationTestUtils.CONTAINER_NAME_0));
                assertNotNull("container 1 does not exist", cfgSvc.getContainerConfiguration(CONTAINER_NAME_1));
                assertEquals("unexpected container number", 2, cfgSvc.getContainersConfiguration().size());

                try {
                    topoSvc.addContainerConfiguration(newContainerConfiguration);
                    fail("Exception " + AddContainerException.class.getName() + " not thrown");
                } catch (final AddContainerException e) {
                    assertEquals("unexpected already existing container", CONTAINER_NAME_1, e.getContainerName());
                }
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

}
