/**
 * Copyright (c) 2014-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import static org.junit.Assert.assertEquals;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.SubDomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerFastRemoteTcpPortException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerHostException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerJmxPortException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerPasswordException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerUserException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentCurrentContainerMovedException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentCurrentContainerNoMoreExistException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentCurrentSubdomainNoMoreExistException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainJndiFactoryException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainJndiProviderUrlException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentSubdomainModeException;
import org.ow2.petals.microkernel.api.configuration.exception.NoTopologyProvidedException;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.Jndi;
import org.ow2.petals.topology.generated.Subdomain;
import org.ow2.petals.topology.generated.SubdomainMode;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.topology.generated.TransportService;

/**
 * Unit tests of {@link ConfigurationServiceImpl#updateTopology(Topology)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ConfigurationServiceImpl_UpdateTopologyTest extends AbstractConfigurationServiceImplTest {

    private final static String DOMAIN_JNDI_FACTORY = "domain.jndi.factory";

    private final static String DOMAIN_JNDI_PROVIDER_URL = "http://domain.jndi.provider.url";

    private final static String SUBDOMAIN_NAME_1 = "subdomain-1";

    private final static SubdomainMode SUBDOMAIN_MODE_1 = SubdomainMode.STATIC;

    private final static String SUBDOMAIN_DESCRIPTION_1 = "A sample sub-domain 1 configuration";

    private final static String CONTAINER_NAME_1 = "sample-1";

    private final static String CONTAINER_DESCRIPTION_1 = "description of the container 1";

    private final static String CONTAINER_NAME_2 = "sample-2";

    private final static String CONTAINER_DESCRIPTION_2 = "description of the container 2";

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>no topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because no topology is provided.
     * </p>
     */
    @Test(expected = NoTopologyProvidedException.class)
    public void testNoTopologyProvided() throws ConfigurationException, InvocationTargetException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.updateTopology(null);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the new topology is the same than the current one</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs</li>.
     * <li>The current topology is always the same.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSameTopology() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);


        final Topology expectedTopology = ConfigurationTestUtils.createBasicSampleTopology();

        try {
            cfgSvc.loadConfiguration(serverProps, expectedTopology);
            assertBasicSampleTopology(cfgSvc);

            cfgSvc.updateTopology(expectedTopology);
            assertBasicSampleTopology(cfgSvc);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the new topology does not contains the current container</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the current container does not exist in the new topology.
     * </p>
     */
    @Test(expected = InconsistentCurrentContainerNoMoreExistException.class)
    public void testCurrentContainerDisappear() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Subdomain finalSubdomain = finalTopology.getDomain().getSubDomain().get(0);
            final Iterator<Container> itFinalContainers = finalSubdomain.getContainer().iterator();
            while (itFinalContainers.hasNext()) {
                final Container finalContainer = itFinalContainers.next();
                if (finalContainer.getName().equals(cfgSvc.getContainerConfiguration().getName())) {
                    itFinalContainers.remove();
                }
            }

            cfgSvc.updateTopology(finalTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the current container has moved in another sub-domain in the new topology,</li>
     * <li>and the current sub-domain no more exists</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the current sub-domain no more exists.
     * </p>
     */
    @Test(expected = InconsistentCurrentSubdomainNoMoreExistException.class)
    public void testCurrentSubdomainDisappear() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Subdomain finalSubdomain = finalTopology.getDomain().getSubDomain().get(0);
            finalSubdomain.setName(SUBDOMAIN_NAME_1);

            cfgSvc.updateTopology(finalTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the current container has moved in another sub-domain in the new topology,</li>
     * <li>and the current sub-domain always exists</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the current container is not in the right sub-domain.
     * </p>
     */
    @Test(expected = InconsistentCurrentContainerMovedException.class)
    public void testCurrentContainerMovedInAnotherSubdomain() throws ConfigurationException, InvocationTargetException,
            IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Domain finalDomain = finalTopology.getDomain();
            final List<Container> containersOfSubdomain0 = finalDomain.getSubDomain().get(0).getContainer();
            final Subdomain finalSubdomain1 = new Subdomain();
            finalDomain.getSubDomain().add(finalSubdomain1);
            finalSubdomain1.setName(SUBDOMAIN_NAME_1);
            finalSubdomain1.setMode(SUBDOMAIN_MODE_1);
            finalSubdomain1.setDescription(SUBDOMAIN_DESCRIPTION_1);
            finalSubdomain1.getContainer().addAll(containersOfSubdomain0);
            containersOfSubdomain0.clear();

            cfgSvc.updateTopology(finalTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the sub-domain mode is changed by the new topology</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the sub-domain mode is changed.
     * </p>
     */
    @Test(expected = InconsistentSubdomainModeException.class)
    public void testSubdomainModeChanged() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            finalTopology.getDomain().getSubDomain().get(0).setMode(SubdomainMode.DYNAMIC);

            cfgSvc.updateTopology(finalTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the jndi factory of the domain is changed by the new topology</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the jndi factory of the domain is changed.
     * </p>
     */
    @Test(expected = InconsistentDomainJndiFactoryException.class)
    public void testDomainJndiFactoryChanged() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Jndi initialJndi = new Jndi();
            initialTopology.getDomain().setJndi(initialJndi);
            initialJndi.setFactory(DOMAIN_JNDI_FACTORY);
            initialJndi.setProviderUrl(DOMAIN_JNDI_PROVIDER_URL);

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Jndi jndi = new Jndi();
            finalTopology.getDomain().setJndi(jndi);
            jndi.setFactory("another." + DOMAIN_JNDI_FACTORY);
            jndi.setProviderUrl(DOMAIN_JNDI_PROVIDER_URL);

            cfgSvc.updateTopology(finalTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the jndi provider url of the domain is changed by the new topology</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the jndi provider url of the domain is changed.
     * </p>
     */
    @Test(expected = InconsistentDomainJndiProviderUrlException.class)
    public void testDomainJndiProviderUrlChanged() throws ConfigurationException, InvocationTargetException,
            IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Jndi initialJndi = new Jndi();
            initialTopology.getDomain().setJndi(initialJndi);
            initialJndi.setFactory(DOMAIN_JNDI_FACTORY);
            initialJndi.setProviderUrl(DOMAIN_JNDI_PROVIDER_URL);

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Jndi jndi = new Jndi();
            finalTopology.getDomain().setJndi(jndi);
            jndi.setFactory(DOMAIN_JNDI_FACTORY);
            jndi.setProviderUrl(DOMAIN_JNDI_PROVIDER_URL + "/another");

            cfgSvc.updateTopology(finalTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the container host is changed by the new topology</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the container host is changed.
     * </p>
     */
    @Test(expected = InconsistentContainerHostException.class)
    public void testContainerHostChanged() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Container finalContainer = finalTopology.getDomain().getSubDomain().get(0).getContainer().get(0);
            finalContainer.setHost("new-host-value");

            cfgSvc.updateTopology(finalTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the container user is changed by the new topology</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the container user is changed.
     * </p>
     */
    @Test(expected = InconsistentContainerUserException.class)
    public void testContainerUserChanged() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Container finalContainer = finalTopology.getDomain().getSubDomain().get(0).getContainer().get(0);
            finalContainer.setUser("new-user-value");

            cfgSvc.updateTopology(finalTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the container password is changed by the new topology</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the container password is changed.
     * </p>
     */
    @Test(expected = InconsistentContainerPasswordException.class)
    public void testContainerPasswordChanged() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Container finalContainer = finalTopology.getDomain().getSubDomain().get(0).getContainer().get(0);
            finalContainer.setPassword("new-password-value");

            cfgSvc.updateTopology(finalTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the container JMX port is changed by the new topology</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the container JMX port is changed.
     * </p>
     */
    @Test(expected = InconsistentContainerJmxPortException.class)
    public void testContainerJmxPortChanged() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Container finalContainer = finalTopology.getDomain().getSubDomain().get(0).getContainer().get(0);
            finalContainer.getJmxService().setRmiPort(9999);

            cfgSvc.updateTopology(finalTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the container fast remote TCP port is changed by the new topology</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the container fast remote TCP port is changed.
     * </p>
     */
    @Test(expected = InconsistentContainerFastRemoteTcpPortException.class)
    public void testContainerFastRemoteTcpPortChanged() throws ConfigurationException, InvocationTargetException,
            IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Container finalContainer = finalTopology.getDomain().getSubDomain().get(0).getContainer().get(0);
            finalContainer.getTransportService().setTcpPort(9999);

            cfgSvc.updateTopology(finalTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the topology that will be updated contains only one container located in one sub-domain,</li>
     * <li>the new topology will:
     * <ul>
     * <li>add a new container in the existing sub-domain,</li>
     * <li>add a new sub-domain with one container.</li>
     * </ul>
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs,</li>
     * <li>the new configuration (ie. topology) is composed of:
     * <ul>
     * <li>one sub-domain (the initial one) with 2 containers (the initial one completed by a new one)</li>
     * <li>another new sub-domain with 1 container</li>
     * </ul>
     * </li>
     * </p>
     */
    @Test
    public void test1NewContainerPlus1SubDomainWith1Container() throws ConfigurationException,
            InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Domain finalDomain = finalTopology.getDomain();
            final Subdomain finalSubdomain0 = finalDomain.getSubDomain().get(0);

            final Container container0_1 = new Container();
            finalSubdomain0.getContainer().add(container0_1);
            container0_1.setName(CONTAINER_NAME_1);
            container0_1.setDescription(CONTAINER_DESCRIPTION_1);
            container0_1.setHost(ConfigurationTestUtils.CONTAINER_HOST);
            container0_1.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
            container0_1.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);

            final JmxService jmxService0_1 = new JmxService();
            container0_1.setJmxService(jmxService0_1);

            final TransportService transportService0_1 = new TransportService();
            transportService0_1.setTcpPort(7781);
            container0_1.setTransportService(transportService0_1);

            final Subdomain finalSubdomain1 = new Subdomain();
            finalDomain.getSubDomain().add(finalSubdomain1);
            finalSubdomain1.setName(SUBDOMAIN_NAME_1);
            finalSubdomain1.setMode(SUBDOMAIN_MODE_1);
            finalSubdomain1.setDescription(SUBDOMAIN_DESCRIPTION_1);

            final Container container1_0 = new Container();
            finalSubdomain1.getContainer().add(container1_0);
            container1_0.setName(CONTAINER_NAME_2);
            container1_0.setDescription(CONTAINER_DESCRIPTION_2);
            container1_0.setHost(ConfigurationTestUtils.CONTAINER_HOST);
            container1_0.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
            container1_0.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);

            final JmxService jmxService1_0 = new JmxService();
            container1_0.setJmxService(jmxService1_0);

            final TransportService transportService1_0 = new TransportService();
            transportService1_0.setTcpPort(7782);
            container1_0.setTransportService(transportService1_0);

            cfgSvc.updateTopology(finalTopology);

            final DomainConfiguration domainConfig = cfgSvc.getDomainConfiguration();
            assertEquals("Unexpected domain name", ConfigurationTestUtils.DOMAIN_NAME, domainConfig.getName());
            assertEquals("Unexpected domain description", ConfigurationTestUtils.DOMAIN_DESCRIPTION,
                    domainConfig.getDescription());
            assertEquals("Unexpected number of sub-domain", 2, cfgSvc.getSubDomainsConfiguration().size());

            final SubDomainConfiguration subdomainConfig = cfgSvc.getSubDomainConfiguration();
            assertEquals("Unexpected sub-domain name", ConfigurationTestUtils.SUBDOMAIN_NAME_0, subdomainConfig.getName());
            assertEquals("Unexpected sub-domain description", ConfigurationTestUtils.SUBDOMAIN_DESCRIPTION_0,
                    subdomainConfig.getDescription());
            assertEquals("Unexpected sub-domain mode", SubDomainConfiguration.SubdomainMode.STATIC,
                    subdomainConfig.getMode());

            assertEquals("Unexpected number of container", 3, cfgSvc.getContainersConfiguration().size());
            final ContainerConfiguration containerConfig = cfgSvc.getContainerConfiguration();
            assertEquals("Unexpected container name", ConfigurationTestUtils.CONTAINER_NAME_0, containerConfig.getName());
            assertEquals("Unexpected container description", ConfigurationTestUtils.CONTAINER_DESCRIPTION_0,
                    containerConfig.getDescription());
            assertEquals("Unexpected container host", ConfigurationTestUtils.CONTAINER_HOST, containerConfig.getHost());
            assertEquals("Unexpected container jmx user", ConfigurationTestUtils.CONTAINER_JMX_USER,
                    containerConfig.getUser());
            assertEquals("Unexpected container jmx pwd", ConfigurationTestUtils.CONTAINER_JMX_PWD,
                    containerConfig.getPassword());

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

}
